fig = Figure()
ax = fig.add_axes([.2,.2, .7, .7])

# Fixing random state for reproducibility
numpy.random.seed(19680801)


N = 50
x = numpy.random.rand(N)
y = numpy.random.rand(N)
colors = numpy.random.rand(N)
area = (30 * numpy.random.rand(N))**2  # 0 to 15 point radii

ax.scatter(x, y, s=area, c=colors, alpha=0.5)

ax.set(title='Bubble chart', xlabel="X", ylabel="Y")

fig



# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------


"""
test_cli
========

Unit tests for cli.py

"""

from argparse import Namespace
from contextlib import contextmanager
from os.path import abspath, dirname, join
import sys
from unittest.mock import patch
from pathlib import Path, PosixPath

import pytest

PYSPREADPATH = abspath(join(dirname(__file__) + "/.."))
LIBPATH = abspath(PYSPREADPATH + "/lib")


@contextmanager
def insert_path(path):
    sys.path.insert(0, path)
    yield
    sys.path.pop(0)


with insert_path(PYSPREADPATH):
    from ..cli import PyspreadArgumentParser

param_test_cli = [
    (['pyspread'],
     Namespace(file=None, default_settings=False, loglevel=30)),
    (['pyspread', 'test.pys'],
     Namespace(file=PosixPath("test.pys"), default_settings=False,
               loglevel=30)),
    (['pyspread', '--help'],
     None),
    (['pyspread', '--version'],
     None),
    (['pyspread', '--default-settings'],
     Namespace(file=None, default_settings=True, loglevel=30)),
    (['pyspread', '-d'],
     Namespace(file=None, default_settings=False, loglevel=10)),
    (['pyspread', '-v'],
     Namespace(file=None, default_settings=False, loglevel=20)),
]


@pytest.mark.parametrize("argv, res", param_test_cli)
def test_cli(argv, res):
    with patch('argparse._sys.argv', argv):
        parser = PyspreadArgumentParser()
        if res is None:
            with pytest.raises(SystemExit) as exc:
                args, unknown = parser.parse_known_args()
            assert exc.value.code == 0
        else:
            args, unknown = parser.parse_known_args()
            assert args == res



# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

"""

**Provides**

 * :class:`MultiStateBitmapButton`
 * :class:`RotationButton`
 * :class:`JustificationButton`
 * :class:`RendererButton`
 * :class:`AlignmentButton`
 * :class:`ColorButton`
 * :class:`TextColorButton`
 * :class:`LineColorButton`
 * :class:`BackgroundColorButton`
 * :class:`MenuComboBox`
 * :class:`TypeMenuComboBox`
 * :class:`FontChoiceCombo`
 * :class:`FontSizeCombo`
 * :class:`Widgets`
 * :class:`FindEditor`
 * :class:`CellButton`
 * :class:`HelpBrowser`

"""

from pathlib import Path
from typing import Tuple

try:
    from markdown2 import markdown
except ImportError:
    markdown = None

from PyQt6.QtCore import pyqtSignal, QSize, Qt, QModelIndex, QPoint
from PyQt6.QtWidgets \
    import (QToolButton, QColorDialog, QFontComboBox, QComboBox, QSizePolicy,
            QLineEdit, QPushButton, QTextBrowser, QWidget, QMainWindow,
            QMenu, QTableView)
from PyQt6.QtGui import (QPalette, QColor, QFont, QIntValidator, QCursor,
                         QIcon, QAction)

try:
    from pyspread.actions import Action
    from pyspread.icons import Icon
    from pyspread.lib.csv import typehandlers, currencies
except ImportError:
    from actions import Action
    from icons import Icon
    from lib.csv import typehandlers, currencies


class MultiStateBitmapButton(QToolButton):
    """QToolButton that cycles through arbitrary states

    The states are defined by an iterable of QIcons

    """

    def __init__(self, main_window: QMainWindow):
        """
        :param main_window: Application main window

        """

        super().__init__()
        self.main_window = main_window

        self._current_action_idx = 0

        self.clicked.connect(self.on_clicked)

    @property
    def current_action_idx(self) -> int:
        """Index of current action"""

        return self._current_action_idx

    @current_action_idx.setter
    def current_action_idx(self, index: int):
        """Sets current action index and updates button and menu

        :param index: Index of action to be set

        """

        self._current_action_idx = index
        action = self.get_action(index)
        self.setIcon(action.icon())

    def get_action(self, index: int) -> QAction:
        """Returns action from index in action_names

        :param index: Index of action to be returned

        """

        action_name = self.action_names[index]
        return self.main_window.main_window_actions[action_name]

    def set_current_action(self, action_name: str):
        """Sets current action

        :param action_name: Name of action as in MainWindowActions

        """

        self.current_action_idx = self.action_names.index(action_name)

    def next(self) -> QAction:
        """Advances current_action_idx and returns current action"""

        if self.current_action_idx >= len(self.action_names) - 1:
            self.current_action_idx = 0
        else:
            self.current_action_idx += 1

        return self.get_action(self.current_action_idx)

    def set_menu_checked(self, action_name: str):
        """Sets checked status of menu

        :param action_name: Name of action as in MainWindowActions

        """

        action = self.main_window.main_window_actions[action_name]
        action.setChecked(True)

    def on_clicked(self):
        """Button clicked event handler. Chechs corresponding menu item"""

        action = self.next()
        action.trigger()
        action.setChecked(True)


class RotationButton(MultiStateBitmapButton):
    """Rotation button for the format toolbar"""

    label = "Rotate"
    action_names = "rotate_0", "rotate_90", "rotate_180", "rotate_270"

    def __init__(self, main_window: QMainWindow):
        """
        :param main_window: Application main window

        """

        super().__init__(main_window)

        self.setStatusTip("Text rotation")
        self.setToolTip("Text rotation")

    def icon(self) -> QIcon:
        """Returns icon for button identification"""

        return Icon.rotate_0


class JustificationButton(MultiStateBitmapButton):
    """Justification button for the format toolbar"""

    label = "Justification"
    action_names = ("justify_left", "justify_center", "justify_right",
                    "justify_fill")

    def __init__(self, main_window: QMainWindow):
        """
        :param main_window: Application main window

        """

        super().__init__(main_window)

        self.setStatusTip("Text justification")
        self.setToolTip("Text justification")

    def icon(self) -> QIcon:
        """Returns icon for button identification"""

        return Icon.justify_left


class RendererButton(MultiStateBitmapButton):
    """Cell render button for the format toolbar"""

    label = "Renderer"
    action_names = "text", "markup", "image", "matplotlib"

    def __init__(self, main_window: QMainWindow):
        """
        :param main_window: Application main window

        """

        super().__init__(main_window)

        self.setStatusTip("Cell render type")
        self.setToolTip("Cell render type")

    def icon(self) -> QIcon:
        """Returns icon for button identification"""

        return Icon.text


class AlignmentButton(MultiStateBitmapButton):
    """Alignment button for the format toolbar"""

    label = "Alignment"
    action_names = "align_top", "align_center", "align_bottom"

    def __init__(self, main_window: QMainWindow):
        """
        :param main_window: Application main window

        """

        super().__init__(main_window)

        self.setStatusTip("Text alignment")
        self.setToolTip("Text alignment")

    def icon(self) -> QIcon:
        """Returns icon for button identification"""

        return Icon.align_top


class ColorButton(QToolButton):
    """Color button widget"""

    colorChanged = pyqtSignal()
    title = "Select Color"
    default_color = None
    _color = None

    def __init__(self, color: QColor, icon: QIcon = None,
                 max_size: QSize = QSize(28, 28)):
        """
        :param color: Color that is initially set
        :param icon: Button foreground image
        :param max_size: Maximum Size of the button

        """

        super().__init__()

        self.set_max_size(max_size)

        if icon is not None:
            self.setIcon(icon)

        self.color = color

        self.pressed.connect(self.on_pressed)

    @property
    def color(self) -> QColor:
        """Chosen color"""

        return self._color

    @color.setter
    def color(self, color: QColor):
        """Color setter that adjusts internal state and button background

        :param color: Color to be set

        """

        if self._color == color:
            return

        self._color = color

        palette = self.palette()
        palette.setColor(QPalette.ColorRole.Button, color)
        self.setAutoFillBackground(True)
        self.setPalette(palette)
        self.update()

    def set_max_size(self, size: QSize):
        """Set the maximum size of the widget

        :param color: Maximum button size

        """

        self.setMaximumWidth(size.width())
        self.setMaximumHeight(size.height())

    def on_pressed(self):
        """Button pressed event handler

        Shows color dialog and sets the chosen color.

        """

        dlg = QColorDialog(self.parent())

        dlg.setCurrentColor(self.color)
        if self.default_color is not None:
            dlg.setCustomColor(15, self.default_color)
        dlg.setWindowTitle(self.title)

        dlg.setWindowFlags(
            Qt.WindowType.Tool | Qt.WindowType.FramelessWindowHint)
        dlg.setWindowModality(Qt.WindowModality.ApplicationModal)
        dlg.setOptions(QColorDialog.ColorDialogOption.DontUseNativeDialog)

        pos = self.mapFromGlobal(QCursor.pos())
        pos.setX(pos.x() + int(self.rect().width() / 2))
        pos.setY(pos.y() + int(self.rect().height() / 2))
        dlg.move(self.mapToGlobal(pos))

        if dlg.exec():
            self.color = dlg.currentColor()
            self.colorChanged.emit()


class TextColorButton(ColorButton):
    """Color button with text icon"""

    label = "Text Color"

    def __init__(self, color: QColor):
        """
        :param color: Color that is initially set

        """

        icon = Icon.text_color
        super().__init__(color, icon=icon)

        self.title = "Select text color"
        self.setStatusTip("Text color")
        self.setToolTip("Text color")

        self.default_color = self.palette().color(QPalette.ColorRole.Text)


class LineColorButton(ColorButton):
    """Color button with text icon"""

    label = "Line Color"

    def __init__(self, color: QColor):
        """
        :param color: Color that is initially set

        """

        icon = Icon.line_color
        super().__init__(color, icon=icon)

        self.title = "Select cell border line color"
        self.setStatusTip("Cell border line color")
        self.setToolTip("Cell border line color")

        self.default_color = self.palette().color(QPalette.ColorRole.Mid)


class BackgroundColorButton(ColorButton):
    """Color button with text icon"""

    label = "Background Color"

    def __init__(self, color: QColor):
        """
        :param color: Color that is initially set

        """

        icon = Icon.background_color
        super().__init__(color, icon=icon)

        self.title = "Select cell background color"
        self.setStatusTip("Cell background color")
        self.setToolTip("Cell background color")

        self.default_color = self.palette().color(QPalette.ColorRole.Base)


class MenuComboBox(QComboBox):
    """ComboBox that uses a menu instead of a list"""

    text_tpl = "{} ({})"  # Template for ComboBox item text

    def __init__(self, items: dict):
        """

        :param items: Menu items

        The dict items needs to be given in the following format:
        {
            "<label_1>": None,  # For leaf item
            "<label_2>": {"<label_2.1>": None},  # For submenu
         }

        """

        super().__init__()

        self.action2index = {}  # Maps menu action to ComboBox item index

        self.menu = QMenu(self)
        self.menu.triggered.connect(self.on_menu_selected)

        self._fill(items, self.menu)

    def _fill(self, items: dict, menu: QMenu, parent_item_text: str = ""):
        """Fills the menu and the combobox

        :param items: Menu items for respective (sub)menu
        :param items: Menu or submenu
        :param parent_item_text: Text of parent item in submenu

        """

        for item in items:
            if parent_item_text:
                text = self.text_tpl.format(parent_item_text, item)
            else:
                text = item

            if items[item] is None:
                # Leaf item
                action = menu.addAction(item)
                self.addItem(text)
                self.action2index[action] = self.count() - 1

            else:
                # Submenu
                submenu = self.menu.addMenu(item)
                self._fill(items[item], submenu, parent_item_text=text)

    def showPopup(self):
        """Show combo menu"""

        rect = self.rect()

        pos = QPoint(rect.x(), rect.y()+rect.height())
        self.menu.popup(self.mapToGlobal(pos))

    def hidePopup(self):
        """Hide combo menu"""

        self.menu.hideTearOffMenu()
        super().hidePopup()

    def on_menu_selected(self, action):
        """Event handler for menu"""

        self.setCurrentIndex(self.action2index[action])


class TypeMenuComboBox(MenuComboBox):
    """MenuComboBox that comprises types and currencies for CSV import"""

    def __init__(self):
        items = {}
        for typehandler in typehandlers:
            if typehandler == "Money":
                items[typehandler] = dict((currency.code, None)
                                          for currency in currencies)
            else:
                items[typehandler] = None

        super().__init__(items)


class FontChoiceCombo(QFontComboBox):
    """Font choice combo box"""

    label = "Font Family"

    fontChanged = pyqtSignal()

    def __init__(self, main_window: QMainWindow):
        """
        :param main_window: Application main window

        """

        super().__init__(main_window)

        self.setMaximumWidth(150)

        # Set default font
        self.setFont(QFont())

        self.currentFontChanged.connect(self.on_font)

    @property
    def font(self) -> str:
        """Font family name"""

        return self.currentFont().family()

    @font.setter
    def font(self, font: str):
        """Sets font from family name without emitting currentTextChanged"""

        self.currentFontChanged.disconnect(self.on_font)
        self.setCurrentFont(QFont(font))
        self.currentFontChanged.connect(self.on_font)

    def icon(self) -> QIcon:
        """Returns QIcon for button identification"""

        return Icon.font_dialog

    def on_font(self):
        """Font choice event handler"""

        self.fontChanged.emit()


class FontSizeCombo(QComboBox):
    """Font choice combo box"""

    label = "Font Size"
    fontSizeChanged = pyqtSignal()

    def __init__(self, main_window: QMainWindow):
        """
        :param main_window: Application main window

        """

        super().__init__()

        self.setEditable(True)

        for size in main_window.settings.font_sizes:
            self.addItem(str(size))

        idx = self.findText(str(main_window.settings.font_sizes))
        if idx >= 0:
            self.setCurrentIndex(idx)

        validator = QIntValidator(1, 128, self)
        self.setValidator(validator)

        self.currentTextChanged.connect(self.on_text)

    @property
    def size(self) -> int:
        """Size of current text"""

        return int(self.currentText())

    @size.setter
    def size(self, size: int):
        """Sets size without emitting currentTextChanged

        :param size: Size to be set

        """

        self.currentTextChanged.disconnect(self.on_text)
        self.setCurrentText(str(size))
        self.currentTextChanged.connect(self.on_text)

    def icon(self) -> QIcon:
        """Returns icon for button identification"""

        return Icon.font_dialog

    def on_text(self):
        """Font size choice event handler"""

        try:
            value = int(self.currentText())
        except ValueError:
            value = 1
            self.setCurrentText("1")

        if value < 1:
            self.setCurrentText("1")
        if value > 128:
            self.setCurrentText("128")

        self.fontSizeChanged.emit()


class Widgets:
    """Container class for widgets"""

    def __init__(self, main_window: QMainWindow):
        """
        :param main_window: Application main window

        """

        # Format toolbar widgets

        self.font_combo = FontChoiceCombo(main_window)

        self.font_size_combo = FontSizeCombo(main_window)

        text_color = QColor("black")
        self.text_color_button = TextColorButton(text_color)

        background_color = QColor("white")
        self.background_color_button = BackgroundColorButton(background_color)

        line_color = QColor("black")
        self.line_color_button = LineColorButton(line_color)

        self.renderer_button = RendererButton(main_window)
        self.rotate_button = RotationButton(main_window)
        self.justify_button = JustificationButton(main_window)
        self.align_button = AlignmentButton(main_window)


class FindEditor(QLineEdit):
    """The Find editor widget for the find toolbar"""

    up = False
    word = False
    case = False
    regexp = False
    results = False

    def __init__(self, parent: QWidget):
        """
        :param parent: Parent widget

        """

        super().__init__(parent)

        self.actions = parent.main_window.main_window_actions

        self.label = "Find editor"
        self.icon = lambda: Icon.find_next
        self.sizePolicy().setHorizontalPolicy(QSizePolicy.Policy.Preferred)
        self.setClearButtonEnabled(True)
        self.addAction(self.actions.find_next,
                       QLineEdit.ActionPosition.LeadingPosition)

        workflows = parent.main_window.workflows
        self.returnPressed.connect(workflows.edit_find_next)

        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self.on_context_menu)

    def prepend_actions(self, menu: QMenu):
        """Prepends find specific actions to menu

        :param menu: Find editor context menu

        """

        toggle_case = Action(self, "Match &case",
                             self.on_toggle_case, checkable=True,
                             statustip='Match case in search')

        toggle_results = Action(self, "Code and results",
                                self.on_toggle_results, checkable=True,
                                statustip='Search also considers string '
                                          'representations of result objects.')

        toggle_up = Action(self, "Search &backward",
                           self.on_toggle_up, checkable=True,
                           statustip='Search fore-/backwards')

        toggle_word = Action(self, "&Whole words",
                             self.on_toggle_word, checkable=True,
                             statustip='Whole word search')

        toggle_regexp = Action(self, "Regular expression",
                               self.on_toggle_regexp, checkable=True,
                               statustip='Regular expression search')

        toggle_case.setChecked(self.case)
        toggle_results.setChecked(self.results)
        toggle_up.setChecked(self.up)
        toggle_word.setChecked(self.word)
        toggle_regexp.setChecked(self.regexp)

        actions = (toggle_case, toggle_results, toggle_up, toggle_word,
                   toggle_regexp)
        menu.insertActions(menu.actions()[0], actions)

    def on_context_menu(self, point: QPoint):
        """Context menu event handler

        :param point: Context menu coordinates on screen

        """

        menu = self.createStandardContextMenu()
        menu.insertSeparator(menu.actions()[0])
        self.prepend_actions(menu)
        menu.exec(self.mapToGlobal(point))

    def on_toggle_up(self, toggled: bool):
        """Find upwards toggle event handler

        :param toggled: up option toggle state

        """

        self.up = toggled

    def on_toggle_word(self, toggled: bool):
        """Find whole word toggle event handler

        :param toggled: whole word option toggle state

        """

        self.word = toggled

    def on_toggle_case(self, toggled: bool):
        """Find case sensitively toggle event handler

        :param toggled: case sensitivity option toggle state

        """

        self.case = toggled

    def on_toggle_regexp(self, toggled: bool):
        """Find with regular expression toggle event handler

        :param toggled: regular expression option toggle state

        """

        self.regexp = toggled

    def on_toggle_results(self, toggled: bool):
        """Find in results toggle event handler

        :param toggled: results option toggle state

        """

        self.results = toggled


class CellButton(QPushButton):
    """Button that is used for button cells in the grid"""

    def __init__(self, text: str, grid: QTableView, key: Tuple[int, int, int]):
        """
        :param text: button label text
        :param grid: Main grid
        :param key: key of button's cell (row, column, table)

        """

        super().__init__(text, grid)

        self.grid = grid
        self.key = key  # Key of button cell

        self.clicked.connect(self.on_clicked)

    def on_clicked(self):
        """Clicked event handler, executes cell code"""

        code = self.grid.model.code_array(self.key)
        result = self.grid.model.code_array._eval_cell(self.key, code)
        self.grid.model.code_array.frozen_cache[repr(self.key)] = result
        self.grid.model.code_array.result_cache.clear()
        self.grid.model.dataChanged.emit(QModelIndex(), QModelIndex())


class HelpBrowser(QTextBrowser):
    """Help browser widget"""

    def __init__(self, parent: QWidget, path: Path):
        """
        :param parent: Parent window
        :param path: Path to markdown file that is displayed

        """

        super().__init__(parent)

        self.setReadOnly(True)
        self.setOpenExternalLinks(True)
        self.update(path)

    def update(self, path: Path):
        """Updates content

        :param path: Path to markdown file that is displayed

        """

        self.setSearchPaths([str(path.parents[0])])
        self.setHtml(self.get_html(path))

    def get_html(self, path: Path) -> str:
        """Returns html content for content of browser

        :param path: Path to markdown file that is displayed

        """

        try:
            with open(path, encoding='utf-8') as helpfile:
                help_text = helpfile.read()
        except IOError as err:
            return "Error opening file {}: {}".format(path, err)

        if markdown is None:
            error_msg = "<b>Warning: markdown2 is not installed.<br>" + \
                        "Rendering as pain text.</b><p>"
            return error_msg + help_text.replace("\n", "<br>")

        return markdown(help_text, extras=['metadata', 'code-friendly',
                                           'fenced-code-blocks'])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------


"""
test_csv
========

Unit tests for csv.py

"""

from csv import QUOTE_NONE
import datetime
from pathlib import Path
import inspect

import pytest

from ..csv import (sniff, get_header, csv_reader, convert, date, time,
                   make_object)
from ..csv import datetime as __datetime


TESTPATH = Path(__file__).parent


param_sniff = [
    (TESTPATH / 'valid1.csv', True, ',', 0, QUOTE_NONE, '"', "\r\n", 0),
    (TESTPATH / 'valid2.csv', True, '\t', 0, QUOTE_NONE, '"', "\r\n", 0),
    (TESTPATH / 'valid3.csv', True, '\t', 0, QUOTE_NONE, '"', "\r\n", 0),
    (TESTPATH / 'valid4.csv', True, ',', 0, QUOTE_NONE, '"', "\r\n", 0),
]


@pytest.mark.parametrize(
    "filepath, hasheader, delimiter, doublequote, quoting, quotechar, "
    "lineterminator, skipinitialspace", param_sniff)
def test_sniff(filepath, hasheader, delimiter, doublequote, quoting, quotechar,
               lineterminator, skipinitialspace):
    """Unit test for sniff"""

    dialect = sniff(filepath, 1024, encoding='utf-8')
    assert dialect.hasheader == hasheader
    assert dialect.delimiter == delimiter
    assert dialect.doublequote == doublequote
    assert dialect.quoting == quoting
    assert dialect.quotechar == quotechar
    assert dialect.lineterminator == lineterminator
    assert dialect.skipinitialspace == skipinitialspace


param_get_header = [
    (TESTPATH / 'valid1.csv', ["a", "b", "c"]),
]


@pytest.mark.parametrize("filepath, header", param_get_header)
def test_get_header(filepath, header):
    """Unit test for get_first_line"""

    dialect = sniff(filepath, 1024, encoding='utf-8')
    dialect.quotechar = None
    with open(filepath) as csvfile:
        __header = get_header(csvfile, dialect)

    assert __header == header


def test_csv_reader():
    """Unit test for csv_reader"""

    filepath = TESTPATH / 'valid1.csv'
    dialect = sniff(filepath, 1024, encoding='utf-8')
    result = [["1", "2", "3"], ["4", "5", "6"]]

    with open(filepath) as csvfile:
        reader = csv_reader(csvfile, dialect)
        for line, resline in zip(reader, result):
            assert line == resline


param_convert = [
    ('12', 'object', '12'),
    ('12', 'str', "'12'"),
    ('12', 'bool', 'True'),
    ('12', 'bytes', "'12'"),
    ('12', 'complex', "(12+0j)"),
    ('12', 'int', "12"),
    ('12', 'float', "12.0"),
    ('12.0', 'repr', "'12.0'"),
    ('12.0', None, "'12.0'"),
    ('12.0', 'object', "12.0"),
    ('2000-1-1', 'date', repr(datetime.date(2000, 1, 1))),
    ('1995-02-05 00:00', 'datetime',
     repr(datetime.datetime(1995, 2, 5, 0, 0))),
    ('23:59:59', 'time', repr(datetime.time(23, 59, 59))),
]


@pytest.mark.parametrize("string, digest_type, res", param_convert)
def test_convert(string, digest_type, res):
    """Unit test for convert"""

    assert convert(string, digest_type) == res


param_date = [
    ("2011-11-1", datetime.date(2011, 11, 1)),
    (42, TypeError),
]


@pytest.mark.parametrize("string, res", param_date)
def test_date(string, res):
    """Unit test for date"""

    if inspect.isclass(res) and issubclass(res, Exception):
        assert isinstance(date(string), res)
    else:
        assert date(string) == res


param_datetime = [
    ("2011-11-1-12:00:00", datetime.datetime(2011, 11, 1, 12, 0, 0)),
    (42, TypeError),
]


param_time = [
    ("12:00", datetime.time(12, 0)),
    (42, TypeError),
]


@pytest.mark.parametrize("string, res", param_time)
def test_time(string, res):
    """Unit test for time"""

    if inspect.isclass(res) and issubclass(res, Exception):
        assert isinstance(time(string), res)
    else:
        assert time(string) == res


@pytest.mark.parametrize("string, res", param_datetime)
def test_datetime(string, res):
    """Unit test for datetime"""

    if inspect.isclass(res) and issubclass(res, Exception):
        assert isinstance(__datetime(string), res)
    else:
        assert __datetime(string) == res


param_make_object = [
    ("42", 42),
    (42, ValueError),
]


@pytest.mark.parametrize("string, res", param_make_object)
def test_make_object(string, res):
    """Unit test for make_object"""

    if inspect.isclass(res) and issubclass(res, Exception):
        with pytest.raises(res):
            make_object(string)
    else:
        assert make_object(string) == res



# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

"""

**Provides**

 * :func:`quote`
 * :func:`wrap_text`

"""

import textwrap

ZEN = """The Zen of Python, by Tim Peters

Beautiful is better than ugly.
Explicit is better than implicit.
Simple is better than complex.
Complex is better than complicated.
Flat is better than nested.
Sparse is better than dense.
Readability counts.
Special cases aren't special enough to break the rules.
Although practicality beats purity.
Errors should never pass silently.
Unless explicitly silenced.
In the face of ambiguity, refuse the temptation to guess.
There should be one-- and preferably only one --obvious way to do it.
Although that way may not be obvious at first unless you're Dutch.
Now is better than never.
Although never is often better than *right* now.
If the implementation is hard to explain, it's a bad idea.
If the implementation is easy to explain, it may be a good idea.
Namespaces are one honking great idea -- let's do more of those!
"""


def quote(code: str) -> str:
    """Quote code

    :param code: Code to be quoted
    :return: Quoted code if not already quoted and quoting possible

    """

    starts_and_ends = [
        ("'", "'"),
        ('"', '"'),
        ("u'", "'"),
        ('u"', '"'),
        ("b'", "'"),
        ('b"', '"'),
        ("r'", "'"),
        ('r"', '"'),
    ]

    if code is None or not (isinstance(code, bytes) or isinstance(code, str)):
        return code

    code = code.strip()

    if code and not (code[0],  code[-1]) in starts_and_ends:
        return repr(code)
    else:
        return code


def wrap_text(text, width=80, maxlen=2000):
    """Wrap text to line width

    :param text: The text to be wrapped
    :param width: Width of the text to be wrapped
    :param maxlen: Maximum total text length before text in truncated and
                   extended by [...]. If None then truncation is disabled.
    :return: Wrapped text

    """

    if text is None:
        return

    if maxlen is not None and len(text) > maxlen:
        text = text[:maxlen] + "..."
    return "\n".join(textwrap.wrap(text, width=width))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------


"""
test_model
==========

Unit tests for model.py

"""
from builtins import zip
from builtins import map
from builtins import str
from builtins import range
from builtins import object

import fractions  # Yes, it is required
import math  # Yes, it is required
from os.path import abspath, dirname, join
import sys

import pytest
import numpy

pyspread_path = abspath(join(dirname(__file__) + "/../.."))
sys.path.insert(0, pyspread_path)

from model.model import (KeyValueStore, CellAttributes, DictGrid, DataArray,
                         CodeArray, CellAttribute, DefaultCellAttributeDict)

from lib.attrdict import AttrDict
from lib.selection import Selection
sys.path.pop(0)


class Settings:
    """Simulates settings class"""

    timeout = 1000


class TestCellAttributes(object):
    """Unit tests for CellAttributes"""

    def setup_method(self, method):
        """Creates empty CellAttributes"""

        self.cell_attr = CellAttributes()

        selection_1 = Selection([(2, 2)], [(4, 5)], [55], [55, 66], [(34, 56)])
        selection_2 = Selection([], [], [], [], [(32, 53), (34, 56)])

        ca1 = CellAttribute(selection_1, 0, AttrDict([("testattr", 3)]))
        ca2 = CellAttribute(selection_2, 0, AttrDict([("testattr", 2)]))

        self.cell_attr.append(ca1)
        self.cell_attr.append(ca2)

    def test_append(self):
        """Test append"""

        selection = Selection([], [], [], [], [(23, 12)])
        table = 0
        attr = AttrDict([("angle", 0.2)])

        self.cell_attr.append(CellAttribute(selection, table, attr))

        # Check if 1 item - the actual action has been added
        assert len(self.cell_attr) == 3

    def test_getitem(self):
        """Test __getitem__"""

        assert self.cell_attr[32, 53, 0].testattr == 2
        assert self.cell_attr[2, 2, 0].testattr == 3

    def test_setitem(self):
        """Test __setitem__"""

        selection_3 = Selection([], [], [], [], [(2, 53), (34, 56)])
        ca3 = CellAttribute(selection_3, 0, AttrDict([("testattr", 5)]))
        self.cell_attr[1] = ca3

        assert not self.cell_attr._attr_cache
        assert not self.cell_attr._table_cache

        assert self.cell_attr[2, 53, 0].testattr == 5

    def test_len_table_cache(self):
        """Test _len_table_cache"""

        self.cell_attr[32, 53, 0]

        assert self.cell_attr._len_table_cache() == 2

    def test_update_table_cache(self):
        """Test _update_table_cache"""

        assert self.cell_attr._len_table_cache() == 0
        self.cell_attr._update_table_cache()
        assert self.cell_attr._len_table_cache() == 2

    def test_get_merging_cell(self):
        """Test get_merging_cell"""

        selection_1 = Selection([(2, 2)], [(5, 5)], [], [], [])
        selection_2 = Selection([(3, 2)], [(9, 9)], [], [], [])
        selection_3 = Selection([(2, 2)], [(9, 9)], [], [], [])

        attr_dict_1 = AttrDict([("merge_area", (2, 2, 5, 5))])
        attr_dict_2 = AttrDict([("merge_area", (3, 2, 9, 9))])
        attr_dict_3 = AttrDict([("merge_area", (2, 2, 9, 9))])

        cell_attribute_1 = CellAttribute(selection_1, 0, attr_dict_1)
        cell_attribute_2 = CellAttribute(selection_2, 0, attr_dict_2)
        cell_attribute_3 = CellAttribute(selection_3, 1, attr_dict_3)

        self.cell_attr.append(cell_attribute_1)
        self.cell_attr.append(cell_attribute_2)
        self.cell_attr.append(cell_attribute_3)

        # Cell 1. 1, 0 is not merged
        assert self.cell_attr.get_merging_cell((1, 1, 0)) is None

        # Cell 3. 3, 0 is merged to cell 3, 2, 0
        assert self.cell_attr.get_merging_cell((3, 3, 0)) == (2, 2, 0)

        # Cell 2. 2, 0 is merged to cell 2, 2, 0
        assert self.cell_attr.get_merging_cell((2, 2, 0)) == (2, 2, 0)

    def test_for_table(self):
        """Test for_table"""

        selection_3 = Selection([], [], [], [], [(2, 53), (34, 56)])
        ca3 = CellAttribute(selection_3, 2, AttrDict([("testattr", 5)]))
        self.cell_attr.append(ca3)

        assert len(self.cell_attr.for_table(0)) == 2

        result_cas = CellAttributes()
        result_cas.append(ca3)
        assert self.cell_attr.for_table(2) == result_cas


class TestKeyValueStore(object):
    """Unit tests for KeyValueStore"""

    def setup_method(self, method):
        """Creates empty KeyValueStore"""

        self.k_v_store = KeyValueStore()

    def test_missing(self):
        """Test if missing value returns None"""

        key = (1, 2, 3)
        assert self.k_v_store[key] is None

        self.k_v_store[key] = 7

        assert self.k_v_store[key] == 7


class TestDictGrid(object):
    """Unit tests for DictGrid"""

    def setup_method(self, method):
        """Creates empty DictGrid"""

        self.dict_grid = DictGrid((100, 100, 100))

    def test_getitem(self):
        """Unit test for __getitem__"""

        with pytest.raises(IndexError):
            self.dict_grid[100, 0, 0]

        self.dict_grid[(2, 4, 5)] = "Test"
        assert self.dict_grid[(2, 4, 5)] == "Test"

    def test_missing(self):
        """Test if missing value returns None"""

        key = (1, 2, 3)
        assert self.dict_grid[key] is None

        self.dict_grid[key] = 7

        assert self.dict_grid[key] == 7


class TestDataArray(object):
    """Unit tests for DataArray"""

    def setup_method(self, method):
        """Creates empty DataArray"""

        self.data_array = DataArray((100, 100, 100), Settings())

    def test_iter(self):
        """Unit test for __iter__"""

        assert list(iter(self.data_array)) == []

        self.data_array[(1, 2, 3)] = "12"
        self.data_array[(1, 2, 4)] = "13"

        assert sorted(list(iter(self.data_array))) == [(1, 2, 3), (1, 2, 4)]

    def test_keys(self):
        """Unit test for keys"""

        assert list(self.data_array.keys()) == []

        self.data_array[(1, 2, 3)] = "12"
        self.data_array[(1, 2, 4)] = "13"

        assert sorted(self.data_array.keys()) == [(1, 2, 3), (1, 2, 4)]

    def test_pop(self):
        """Unit test for pop"""

        self.data_array[(1, 2, 3)] = "12"
        self.data_array[(1, 2, 4)] = "13"

        assert self.data_array.pop((1, 2, 3)) == "12"

        assert sorted(self.data_array.keys()) == [(1, 2, 4)]

    def test_get_shape(self):
        """Unit test for _get_shape"""

        assert self.data_array.shape == (100, 100, 100)

    def test_set_shape(self):
        """Unit test for _set_shape"""

        self.data_array.shape = (10000, 100, 100)
        assert self.data_array.shape == (10000, 100, 100)

    param_get_last_filled_cell = [
        ({(0, 0, 0): "2"}, 0, (0, 0)),
        ({(2, 0, 2): "2"}, 0, (0, 0)),
        ({(2, 0, 2): "2"}, None, (2, 0)),
        ({(2, 0, 2): "2"}, 2, (2, 0)),
        ({(32, 30, 0): "432"}, 0, (32, 30)),
    ]

    @pytest.mark.parametrize("content,table,res", param_get_last_filled_cell)
    def test_get_last_filled_cell(self, content, table, res):
        """Unit test for get_last_filled_cellet_end"""

        for key in content:
            self.data_array[key] = content[key]

        assert self.data_array.get_last_filled_cell(table)[:2] == res

    def test_getstate(self):
        """Unit test for __getstate__ (pickle support)"""

        assert "dict_grid" in self.data_array.__getstate__()

    def test_slicing(self):
        """Unit test for __getitem__ and __setitem__"""

        self.data_array[0, 0, 0] = "'Test'"
        self.data_array[0, 0, 0] = "'Tes'"

        assert self.data_array[0, 0, 0] == "'Tes'"

    def test_cell_array_generator(self):
        """Unit test for cell_array_generator"""

        cell_array = self.data_array[:5, 0, 0]

        assert list(cell_array) == [None] * 5

        cell_array = self.data_array[:5, :5, 0]

        assert [list(c) for c in cell_array] == [[None] * 5] * 5

        cell_array = self.data_array[:5, :5, :5]

        assert [[list(e) for e in c] for c in cell_array] == \
            [[[None] * 5] * 5] * 5

    param_adjust_rowcol = [
        ({(0, 0): 3.0}, 0, 2, 0, 0, (0, 0), 3.0),
        ({(0, 0): 3.0}, 0, 2, 0, 0, (2, 0), 3.0),
        ({(0, 0): 3.0}, 0, 1, 1, 0, (1, 0), 3.0),
        ({(0, 0): 3.0}, 0, 1, 1, 0, (0, 1), 0.0),
    ]

    @pytest.mark.parametrize("vals, ins_point, no2ins, axis, tab, target, res",
                             param_adjust_rowcol)
    def test_adjust_rowcol(self, vals, ins_point, no2ins, axis, tab, target,
                           res):
        """Unit test for _adjust_rowcol"""

        if axis == 0:
            __vals = self.data_array.row_heights
        elif axis == 1:
            __vals = self.data_array.col_widths
        else:
            raise ValueError("{} out of 0, 1".format(axis))

        __vals.update(vals)

        self.data_array._adjust_rowcol(ins_point, no2ins, axis, tab)
        assert __vals[target] == res

    def test_set_cell_attributes(self):
        """Unit test for _set_cell_attributes"""

        cell_attributes = self.data_array.cell_attributes

        attr = CellAttribute(Selection([], [], [], [], []), 0,
                             AttrDict([("Test", None)]))
        cell_attributes.clear()
        cell_attributes.append(attr)

        assert self.data_array.cell_attributes == cell_attributes

    param_adjust_cell_attributes = [
        (0, 5, 0, (4, 3, 0), (9, 3, 0)),
        (34, 5, 0, (4, 3, 0), (4, 3, 0)),
        (0, 0, 0, (4, 3, 0), (4, 3, 0)),
        (1, 5, 1, (4, 3, 0), (4, 8, 0)),
        (1, 5, 1, (4, 3, 1), (4, 8, 1)),
        (0, -1, 2, (4, 3, 1), None),
        (0, -1, 2, (4, 3, 2), (4, 3, 1)),
    ]

    @pytest.mark.parametrize("inspoint, noins, axis, src, target",
                             param_adjust_cell_attributes)
    def test_adjust_cell_attributes(self, inspoint, noins, axis, src, target):
        """Unit test for _adjust_cell_attributes"""

        row, col, tab = src

        cell_attributes = self.data_array.cell_attributes

        attr_dict = AttrDict([("angle", 0.2)])
        attr = CellAttribute(Selection([], [], [], [], [(row, col)]), tab,
                             attr_dict)
        cell_attributes.clear()
        cell_attributes.append(attr)

        self.data_array._adjust_cell_attributes(inspoint, noins, axis)

        if target is None:
            for key in attr_dict:
                # Should be at default value
                default_ca = DefaultCellAttributeDict()[key]
                assert cell_attributes[src][key] == default_ca
        else:
            for key in attr_dict:
                assert cell_attributes[target][key] == attr_dict[key]

    param_test_insert = [
        ({(2, 3, 0): "42"}, 1, 1, 0, None,
         {(2, 3, 0): None, (3, 3, 0): "42"}),
        ({(0, 0, 0): "0", (0, 0, 2): "2"}, 1, 1, 2, None,
         {(0, 0, 3): "2", (0, 0, 4): None}),
    ]

    @pytest.mark.parametrize("data, inspoint, notoins, axis, tab, res",
                             param_test_insert)
    def test_insert(self, data, inspoint, notoins, axis, tab, res):
        """Unit test for insert operation"""

        self.data_array.dict_grid.update(data)
        self.data_array.insert(inspoint, notoins, axis, tab)

        for key in res:
            assert self.data_array[key] == res[key]

    param_test_delete = [
        ({(2, 3, 4): "42"}, 1, 1, 0, None, {(1, 3, 4): "42"}),
        ({(0, 0, 0): "1"}, 0, 1, 0, 0, {(0, 0, 0): None}),
        ({(0, 0, 1): "1"}, 0, 1, 2, None, {(0, 0, 0): "1"}),
        ({(3, 3, 2): "3"}, 0, 2, 2, None, {(3, 3, 0): "3"}),
        ({(4, 2, 1): "3"}, 2, 1, 1, 1, {(4, 2, 1): None}),
        ({(10, 0, 0): "1"}, 0, 10, 0, 0, {(0, 0, 0): "1"}),
    ]

    @pytest.mark.parametrize("data, delpoint, notodel, axis, tab, res",
                             param_test_delete)
    def test_delete(self, data, delpoint, notodel, axis, tab, res):
        """Tests delete operation"""

        self.data_array.dict_grid.update(data)
        self.data_array.delete(delpoint, notodel, axis, tab)

        for key in res:
            assert self.data_array[key] == res[key]

    def test_delete_error(self):
        """Tests delete operation error"""

        self.data_array[2, 3, 4] = "42"

        try:
            self.data_array.delete(1, 1, 20)
            assert False
        except ValueError:
            pass

    def test_set_row_height(self):
        """Unit test for set_row_height"""

        self.data_array.set_row_height(7, 1, 22.345)
        assert self.data_array.row_heights[7, 1] == 22.345

    def test_set_col_width(self):
        """Unit test for set_col_width"""

        self.data_array.set_col_width(7, 1, 22.345)
        assert self.data_array.col_widths[7, 1] == 22.345


class TestCodeArray(object):
    """Unit tests for CodeArray"""

    def setup_method(self, method):
        """Creates empty DataArray"""

        self.code_array = CodeArray((100, 10, 3), Settings())

    param_test_setitem = [
        ({(2, 3, 2): "42"}, {(1, 3, 2): "42"},
         {(1, 3, 2): "42", (2, 3, 2): "42"}),
    ]

    @pytest.mark.parametrize("data, items, res_data", param_test_setitem)
    def test_setitem(self, data, items, res_data):
        """Unit test for __setitem__"""

        self.code_array.dict_grid.update(data)
        for key in items:
            self.code_array[key] = items[key]
        for key in res_data:
            assert res_data[key] == self.code_array(key)

    def test_slicing(self):
        """Unit test for __getitem__ and __setitem__"""

        # Test for item getting, slicing, basic evaluation correctness

        shape = self.code_array.shape
        x_list = [0, shape[0]-1]
        y_list = [0, shape[1]-1]
        z_list = [0, shape[2]-1]
        for x, y, z in zip(x_list, y_list, z_list):
            assert self.code_array[x, y, z] is None
            self.code_array[:x, :y, :z]
            self.code_array[:x:2, :y:2, :z:-1]

        get_shape = numpy.array(self.code_array[:, :, :]).shape
        orig_shape = self.code_array.shape
        assert get_shape == orig_shape

        gridsize = 100
        filled_grid = CodeArray((gridsize, 10, 1), Settings())
        for i in [-2**99, 2**99, 0]:
            for j in range(gridsize):
                filled_grid[j, 0, 0] = str(i)
                filled_grid[j, 1, 0] = str(i) + '+' + str(j)
                filled_grid[j, 2, 0] = str(i) + '*' + str(j)

            for j in range(gridsize):
                assert filled_grid[j, 0, 0] == i
                assert filled_grid[j, 1, 0] == i + j
                assert filled_grid[j, 2, 0] == i * j

            for j, funcname in enumerate(['int', 'math.ceil',
                                          'fractions.Fraction']):
                filled_grid[0, 0, 0] = "fractions = __import__('fractions')"
                filled_grid[0, 0, 0]
                filled_grid[1, 0, 0] = "math = __import__('math')"
                filled_grid[1, 0, 0]
                filled_grid[j, 3, 0] = funcname + ' (' + str(i) + ')'

                assert filled_grid[j, 3, 0] == eval(funcname + "(" + "i" + ")")
        # Test X, Y, Z
        for i in range(10):
            self.code_array[i, 0, 0] = str(i)
        assert [self.code_array((i, 0, 0)) for i in range(10)] == \
            list(map(str, range(10)))

        assert [self.code_array[i, 0, 0] for i in range(10)] == list(range(10))

        # Test cycle detection

        filled_grid[0, 0, 0] = "numpy.arange(0, 10, 0.1)"
        filled_grid[1, 0, 0] = "sum(S[0,0,0])"

        assert filled_grid[1, 0, 0] == sum(numpy.arange(0, 10, 0.1))

    def test_make_nested_list(self):
        """Unit test for _make_nested_list"""

        def gen():
            """Nested generator"""

            yield (("Test" for _ in range(2)) for _ in range(2))

        res = self.code_array._make_nested_list(gen())

        assert res == [[["Test" for _ in range(2)] for _ in range(2)]]

    data_eval_cell = [
        ((0, 0, 0), "2 + 4", 6),
        ((1, 0, 0), "S[0, 0, 0]", None),
        ((43, 2, 1), "X, Y, Z", (43, 2, 1)),
    ]

    @pytest.mark.parametrize("key, code, res", data_eval_cell)
    def test_eval_cell(self, key, code, res):
        """Unit test for _eval_cell"""

        self.code_array[key] = code
        assert self.code_array._eval_cell(key, code) == res

    def test_execute_macros(self):
        """Unit test for execute_macros"""

        self.code_array.macros = "a = 5\ndef f(x): return x ** 2"
        self.code_array.execute_macros()
        assert self.code_array._eval_cell((0, 0, 0), "a") == 5
        assert self.code_array._eval_cell((0, 0, 0), "f(2)") == 4

    def test_sorted_keys(self):
        """Unit test for _sorted_keys"""

        code_array = self.code_array

        keys = [(1, 0, 0), (2, 0, 0), (0, 1, 0), (0, 99, 0), (0, 0, 0),
                (0, 0, 99), (1, 2, 3)]
        sorted_keys = [(0, 1, 0), (0, 99, 0), (1, 2, 3), (0, 0, 99), (0, 0, 0),
                       (1, 0, 0), (2, 0, 0)]
        rev_sorted_keys = [(0, 1, 0), (2, 0, 0), (1, 0, 0), (0, 0, 0),
                           (0, 0, 99), (1, 2, 3), (0, 99, 0)]

        sort_gen = code_array._sorted_keys(keys, (0, 1, 0))
        for result, expected_result in zip(sort_gen, sorted_keys):
            assert result == expected_result

        rev_sort_gen = code_array._sorted_keys(keys, (0, 3, 0), reverse=True)
        for result, expected_result in zip(rev_sort_gen, rev_sorted_keys):
            assert result == expected_result

    def test_string_match(self):
        """Tests creation of string_match"""

        code_array = self.code_array

        test_strings = [
            "", "Hello", " Hello", "Hello ", " Hello ", "Hello\n",
            "THelloT", " HelloT", "THello ", "hello", "HELLO", "sd"
        ]

        search_string = "Hello"

        # Normal search
        flags = False, False, False
        results = [None, 0, 1, 0, 1, 0, 1, 1, 1, 0, 0, None]
        for test_string, result in zip(test_strings, results):
            res = code_array.string_match(test_string, search_string, *flags)
            assert res == result

        # Case sensitive
        flags = False, True, False
        results = [None, 0, 1, 0, 1, 0, 1, 1, 1, None, None, None]
        for test_string, result in zip(test_strings, results):
            res = code_array.string_match(test_string, search_string, *flags)
            assert res == result

        # Word search
        flags = True, False, False
        results = [None, 0, 1, 0, 1, 0, None, None, None, 0, 0, None]
        for test_string, result in zip(test_strings, results):
            res = code_array.string_match(test_string, search_string, *flags)
            assert res == result

    def test_findnextmatch(self):
        """Find method test"""

        code_array = self.code_array

        for i in range(100):
            code_array[i, 0, 0] = str(i)

        assert code_array[3, 0, 0] == 3
        assert code_array.findnextmatch((0, 0, 0), "3", False) == (3, 0, 0)
        assert code_array.findnextmatch((0, 0, 0), "99", True) == (99, 0, 0)



fig = Figure()
ax = fig.add_axes([.2, .2, .7, .7])

labels = ['G1', 'G2', 'G3', 'G4', 'G5']
men_means = [20, 34, 30, 35, 27]
women_means = [25, 32, 34, 20, 25]

x = numpy.arange(len(labels))  # the label locations
width = 0.35  # the width of the bars

rects1 = ax.bar(x - width/2, men_means, width, label='Men')
rects2 = ax.bar(x + width/2, women_means, width, label='Women')

# Add some text for labels, title and custom x-axis tick labels, etc.
ax.set_ylabel('Scores')
ax.set_title('Grouped column chart')
ax.set_xticks(x)
ax.set_xticklabels(labels)
ax.set_ylim(None, 50)
ax.legend()


def autolabel(rects, ax):
    """Attach a text label above each bar in *rects*, displaying its height."""
    for rect in rects:
        height = rect.get_height()
        ax.annotate('{}'.format(height),
                    xy=(rect.get_x() + rect.get_width() / 2, height),
                    xytext=(0, 3),  # 3 points vertical offset
                    textcoords="offset points",
                    ha='center', va='bottom')


autolabel(rects1, ax)
autolabel(rects2, ax)

fig


from matplotlib import cm

fig = Figure()
ax = fig.add_axes([.2,.2, .7, .7])

size = 0.3
vals = numpy.array([[60., 32.], [37., 40.], [29., 10.]])

cmap = cm.tab20c
outer_colors = cmap(numpy.arange(3)*4)
inner_colors = cmap(numpy.array([1, 2, 5, 6, 9, 10]))

ax.pie(vals.sum(axis=1), radius=1, colors=outer_colors,
       wedgeprops=dict(width=size, edgecolor='w'))

ax.pie(vals.flatten(), radius=1-size, colors=inner_colors,
       wedgeprops=dict(width=size, edgecolor='w'))

ax.set(aspect="equal", title="Pie ring chart")

fig



# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

"""

**Provides**

* :class:`MenuBar`: QMenuBar, the main menubar
* :class:`FileMenu`: File menu for the main menubar
* :class:`EditMenu`: Edit menu for the main menubar
* :class:`ViewMenu`: View menu for the main menubar
* :class:`FormatMenu`: Format menu for the main menubar
* :class:`MacroMenu`: Macro menu for the main menubar
* :class:`HelpMenu`: Help menu for the main menubar
* :class:`FileHistoryMenu`: Menu showing recent files
* :class:`BorderChoiceMenu`: QMenu for choosing cell borders
* :class:`BorderWidthMenu`: QMenu for choosing the cell border width

"""

from functools import partial
from pathlib import Path

from PyQt6.QtCore import Qt
from PyQt6.QtGui import QAction
from PyQt6.QtWidgets import QMenuBar, QMenu, QMainWindow, QWidget, QToolBar

try:
    import matplotlib.figure as matplotlib_figure
except ImportError:
    matplotlib_figure = None

try:
    import dateutil
except ImportError:
    dateutil = None

try:
    from pyspread.actions import MainWindowActions
    from pyspread.icons import Icon
except ImportError:
    from actions import MainWindowActions
    from icons import Icon


class MenuBar(QMenuBar):
    """The main menubar """

    def __init__(self, main_window: QMainWindow):
        """
        :param main_window: Application main window

        """

        super().__init__()

        self.main_window = main_window
        actions = main_window.main_window_actions

        self.file_menu = FileMenu(self, actions)
        self.edit_menu = EditMenu(self, actions)
        self.view_menu = ViewMenu(self, actions)
        self.format_menu = FormatMenu(self, actions)
        self.macro_menu = MacroMenu(self, actions)
        self.help_menu = HelpMenu(self, actions)

        self.addMenu(self.file_menu)
        self.addMenu(self.edit_menu)
        self.addMenu(self.view_menu)
        self.addMenu(self.format_menu)
        self.addMenu(self.macro_menu)
        self.addMenu(self.help_menu)


class FileMenu(QMenu):
    """File menu for the main menubar"""

    def __init__(self, parent: QWidget, actions: MainWindowActions):
        """
        :param parent: Parent widget
        :param actions: Main window actions

        """

        super().__init__('&File', parent)

        self.parent = parent

        self.addAction(actions.new)
        self.addAction(actions.open)
        self.history_submenu = FileHistoryMenu(self, actions)
        self.history_action = self.addMenu(self.history_submenu)
        self.addSeparator()
        self.addAction(actions.save)
        self.addAction(actions.save_as)
        self.addSeparator()
        self.addAction(actions.imprt)
        self.addAction(actions.export)
        self.addSeparator()
        self.addAction(actions.approve)
        self.addSeparator()
        self.addAction(actions.clear_globals)
        self.addSeparator()
        self.addAction(actions.print_preview)
        self.addAction(actions.print)
        self.addSeparator()
        self.addAction(actions.preferences)
        self.addSeparator()
        self.addAction(actions.quit)


class EditMenu(QMenu):
    """Edit menu for the main menubar"""

    def __init__(self, parent: QWidget, actions: MainWindowActions):
        """
        :param parent: Parent widget
        :param actions: Main window actions

        """

        super().__init__('&Edit', parent)

        self.addAction(actions.undo)
        self.addAction(actions.redo)
        self.addSeparator()
        self.addAction(actions.cut)
        self.addAction(actions.copy)
        self.addAction(actions.copy_results)
        self.addAction(actions.paste)
        self.addAction(actions.paste_as)
        self.addSeparator()
        self.addAction(actions.find)
        self.addAction(actions.replace)
        self.addSeparator()
        self.addAction(actions.sort_ascending)
        self.addAction(actions.sort_descending)
        self.addSeparator()
        self.addAction(actions.toggle_selection_mode)
        self.addSeparator()
        self.addAction(actions.shift_cells_down)
        self.addAction(actions.shift_cells_right)
        self.addAction(actions.delete_shift_cells_up)
        self.addAction(actions.delete_shift_cells_left)
        self.addSeparator()
        self.addAction(actions.insert_rows)
        self.addAction(actions.insert_columns)
        self.addAction(actions.insert_table)
        self.addSeparator()
        self.addAction(actions.delete_rows)
        self.addAction(actions.delete_columns)
        self.addAction(actions.delete_table)
        self.addSeparator()
        self.addAction(actions.resize_grid)


class ViewMenu(QMenu):
    """View menu for the main menubar"""

    def __init__(self, parent: QWidget, actions: MainWindowActions):
        """
        :param parent: Parent widget
        :param actions: Main window actions

        """

        super().__init__('&View', parent)

        self.addAction(actions.fullscreen)
        self.addSeparator()

        self.toolbar_submenu = self.addMenu('Toolbars')
        self.toolbar_submenu.addAction(actions.toggle_main_toolbar)
        self.toolbar_submenu.addAction(actions.toggle_macro_toolbar)
        self.toolbar_submenu.addAction(actions.toggle_format_toolbar)
        self.toolbar_submenu.addAction(actions.toggle_find_toolbar)

        self.addAction(actions.toggle_entry_line_dock)
        self.addAction(actions.toggle_macro_dock)
        self.addSeparator()
        self.addAction(actions.goto_cell)
        self.addSeparator()
        self.addAction(actions.toggle_spell_checker)
        self.addSeparator()
        self.addAction(actions.zoom_in)
        self.addAction(actions.zoom_out)
        self.addAction(actions.zoom_1)
        self.addSeparator()
        self.addAction(actions.refresh_cells)
        self.addAction(actions.toggle_periodic_updates)
        self.addSeparator()
        self.addAction(actions.show_frozen)


class FormatMenu(QMenu):
    """Format menu for the main menubar"""

    def __init__(self, parent: QWidget, actions: MainWindowActions):
        """
        :param parent: Parent widget
        :param actions: Main window actions

        """

        super().__init__('&Format', parent)

        self.addAction(actions.copy_format)
        self.addAction(actions.paste_format)
        self.addSeparator()
        self.addAction(actions.default_format)
        self.addSeparator()
        self.addAction(actions.font)
        self.addAction(actions.bold)
        self.addAction(actions.italics)
        self.addAction(actions.underline)
        self.addAction(actions.strikethrough)
        self.addSeparator()

        self.renderer_submenu = self.addMenu('Cell renderer')
        self.renderer_submenu.addAction(actions.text)
        self.renderer_submenu.addAction(actions.image)
        self.renderer_submenu.addAction(actions.markup)
        if matplotlib_figure is not None:
            self.renderer_submenu.addAction(actions.matplotlib)

        self.addAction(actions.freeze_cell)
        self.addAction(actions.lock_cell)
        self.addAction(actions.button_cell)

        self.addSeparator()

        self.addAction(actions.merge_cells)

        self.addSeparator()

        self.rotation_submenu = self.addMenu('Rotation')
        self.rotation_submenu.addAction(actions.rotate_0)
        self.rotation_submenu.addAction(actions.rotate_90)
        self.rotation_submenu.addAction(actions.rotate_180)
        self.rotation_submenu.addAction(actions.rotate_270)

        self.justification_submenu = self.addMenu('Justification')
        self.justification_submenu.addAction(actions.justify_left)
        self.justification_submenu.addAction(actions.justify_center)
        self.justification_submenu.addAction(actions.justify_right)
        self.justification_submenu.addAction(actions.justify_fill)

        self.alignment_submenu = self.addMenu('Alignment')
        self.alignment_submenu.addAction(actions.align_top)
        self.alignment_submenu.addAction(actions.align_center)
        self.alignment_submenu.addAction(actions.align_bottom)

        self.addSeparator()

        self.border_submenu = BorderChoiceMenu(actions)
        self.addMenu(self.border_submenu)

        self.line_width_submenu = BorderWidthMenu(actions)
        self.addMenu(self.line_width_submenu)

        self.addSeparator()
        self.addAction(actions.text_color)
        self.addAction(actions.line_color)
        self.addAction(actions.background_color)


class MacroMenu(QMenu):
    """Macro menu for the main menubar"""

    def __init__(self, parent: QWidget, actions: MainWindowActions):
        """
        :param parent: Parent widget
        :param actions: Main window actions

        """

        super().__init__('&Macro', parent)

        self.addAction(actions.insert_image)
        if matplotlib_figure is not None:
            self.addAction(actions.insert_chart)
        self.addSeparator()
        self.addAction(actions.quote)
        self.addAction(actions.money)
        if dateutil is not None:
            self.addAction(actions.datetime)
            self.addAction(actions.date)
            self.addAction(actions.time)
        self.addSeparator()
        self.addAction(actions.insert_sum)


class HelpMenu(QMenu):
    """Help menu for the main menubar"""

    def __init__(self, parent: QWidget, actions: MainWindowActions):
        """
        :param parent: Parent widget
        :param actions: Main window actions

        """

        super().__init__('&Help', parent)

        self.addAction(actions.manual)
        self.addAction(actions.tutorial)
        self.addSeparator()
        self.addAction(actions.dependencies)
        self.addSeparator()
        self.addAction(actions.about)


class FileHistoryMenu(QMenu):
    """Menu that displays the file history"""

    def __init__(self, parent: QWidget, actions: MainWindowActions):
        """
        :param parent: Parent widget
        :param actions: Main window actions

        """

        super().__init__('&Recent files', parent)

        self.main_window = parent.parent.main_window

    def update_(self):
        """Updates file history menu"""

        self.clear()

        settings = self.main_window.settings
        for posixpath in settings.file_history[:settings.max_file_history]:
            filepath = Path(posixpath)
            if filepath.is_file():
                action = QAction(filepath.name, self)
                action.setStatusTip(posixpath)
                self.addAction(action)
                action.triggered.connect(self.on_recent)

    def on_recent(self):
        """Event handler for file history menu"""

        posixpath = self.sender().statusTip()
        self.main_window.workflows.file_open_recent(posixpath)


class BorderChoiceMenu(QMenu):
    """QMenu for choosing cell borders that shall be manipulated"""

    def __init__(self, actions: MainWindowActions):
        """
        :param actions: Main window actions

        """

        self.actions = actions

        super().__init__()

        self.setTitle("Formatted borders")
        self.setIcon(Icon.border_menu)

        self.addAction(actions.format_borders_all)
        self.addAction(actions.format_borders_top)
        self.addAction(actions.format_borders_bottom)
        self.addAction(actions.format_borders_left)
        self.addAction(actions.format_borders_right)
        self.addAction(actions.format_borders_outer)
        self.addAction(actions.format_borders_inner)
        self.addAction(actions.format_borders_top_bottom)

        self.triggered.connect(self.on_selection)

    def on_selection(self, event):
        """Event handler"""

        self.actions.parent.format_toolbar.border_menu_button.setAttribute(
            Qt.WA_UnderMouse, False)


class BorderWidthMenu(QMenu):
    """QMenu for choosing the cell border width"""

    def __init__(self, actions: MainWindowActions):
        """
        :param actions: Main window actions

        """

        self.actions = actions

        super().__init__()

        self.setTitle("Border width")
        self.setIcon(Icon.format_borders)

        self.addAction(actions.format_borders_0)
        self.addAction(actions.format_borders_1)
        self.addAction(actions.format_borders_2)
        self.addAction(actions.format_borders_4)
        self.addAction(actions.format_borders_8)
        self.addAction(actions.format_borders_16)
        self.addAction(actions.format_borders_32)
        self.addAction(actions.format_borders_64)

        self.triggered.connect(self.on_selection)

    def on_selection(self, event):
        """Event handler"""

        self.actions.parent.format_toolbar.border_menu_button.setAttribute(
            Qt.WidgetAttribute.WA_UnderMouse, False)


class GridContextMenu(QMenu):
    """Context menu for grid"""

    def __init__(self, actions: MainWindowActions):
        """
        :param actions: Main window actions

        """

        super().__init__()

        self.addAction(actions.cut)
        self.addAction(actions.copy)
        self.addAction(actions.copy_results)
        self.addAction(actions.copy_format)
        self.addAction(actions.paste)
        self.addAction(actions.paste_as)
        self.addAction(actions.paste_format)
        self.addSeparator()
        self.addAction(actions.toggle_selection_mode)
        self.addSeparator()
        self.addAction(actions.copy_format)
        self.addAction(actions.paste_format)
        self.addSeparator()
        self.addAction(actions.toggle_selection_mode)
        self.addSeparator()
        self.addAction(actions.quote)
        self.addSeparator()
        self.addAction(actions.shift_cells_down)
        self.addAction(actions.shift_cells_right)
        self.addAction(actions.delete_shift_cells_up)
        self.addAction(actions.delete_shift_cells_left)
        self.addSeparator()
        self.addAction(actions.insert_rows)
        self.addAction(actions.insert_columns)
        self.addSeparator()
        self.addAction(actions.delete_rows)
        self.addAction(actions.delete_columns)


class HorizontalHeaderContextMenu(QMenu):
    """Context menu for horizontal grid header"""

    def __init__(self, actions: MainWindowActions):
        """
        :param actions: Main window actions

        """

        super().__init__()

        self.addAction(actions.insert_columns)
        self.addAction(actions.delete_columns)


class VerticalHeaderContextMenu(QMenu):
    """Context menu for vertical grid header"""

    def __init__(self, actions: MainWindowActions):
        """
        :param actions: Main window actions

        """

        super().__init__()

        self.addAction(actions.insert_rows)
        self.addAction(actions.delete_rows)


class TableChoiceContextMenu(QMenu):
    """Context menu for table choice"""

    def __init__(self, actions: MainWindowActions):
        """
        :param actions: Main window actions

        """

        super().__init__()

        self.addAction(actions.insert_table)
        self.addAction(actions.delete_table)


class ToolbarManagerMenu(QMenu):
    """Menu with all actions of a toolbar that allows toggling visibility"""

    def __init__(self, toolbar: QToolBar):
        """
        :param toolbar: Toolbar that the menu is managing

        """

        super().__init__()
        self.toolbar = toolbar

        for action in toolbar.actions():
            if action.isSeparator():
                self.addSeparator()
            else:
                self.addAction(self._get_toggle_action(action))
        self.update_checked_states()

    def _get_toggle_action(self, action: QAction) -> QAction:
        """Returns a toggle action for a toolbar action

        :param action: Toolbar action for which the toggle state is returned

        """

        taction = QAction(action.icon(), action.text(), action, checkable=True)
        taction.triggered.connect(partial(self.on_toggled, action))

        return taction

    def update_checked_states(self):
        """Updates checked states"""

        for tool_action, action in zip(self.toolbar.actions(), self.actions()):
            action.setChecked(tool_action.isVisible())

    def on_toggled(self, action: QAction, toggled: bool):
        """Action toggle event handler

        :param action: Toolbar action for which visibility is toggled
        :param toggled: If False then action is set invisible and vice versa

        """

        action.setVisible(toggled)



# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

"""

**Provides**

 * :class:`ProgressDialogCanceled`
 * :func:`progress_dialog`
 * :func:`linecount`
 * :func:`file_progress_gen`

"""

from contextlib import contextmanager
from functools import partial
from typing import BinaryIO, ContextManager, Generator, IO, Tuple

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QApplication, QMainWindow, QProgressDialog


class ProgressDialogCanceled(Exception):
    """Raised when a progress dialog is canceled"""

    pass


@contextmanager
def progress_dialog(main_window: QMainWindow, title: str, label: str,
                    maximum: int) -> ContextManager[QProgressDialog]:
    """:class:`~contextlib.contextmanager` that displays a progress dialog

    :param main_window: Application main window
    :param title: Progress dialog title
    :param label: Progress dialog label
    :param maximum: Maximum value for progress

    """

    progress_dialog = QProgressDialog(main_window)
    progress_dialog.setWindowTitle(title)
    progress_dialog.setWindowModality(Qt.WindowModality.WindowModal)
    progress_dialog.setLabelText(label)
    progress_dialog.setMaximum(maximum)

    yield progress_dialog

    progress_dialog.setValue(maximum)
    progress_dialog.close()
    progress_dialog.deleteLater()


def linecount(infile: BinaryIO, buffer_size: int = 1024*1024) -> int:
    """Count lines in infile

    Starts at current position in file. Position is not preserved.
    Idea taken from https://stackoverflow.com/questions/845058

    :param infile: File like object for which lines are counted (binary mode!)
    :param buffer_size: Buffer size for reading the file
    :return: Number of newlines in infile

    """

    buffer_gen = iter(partial(infile.raw.read, buffer_size), b'')
    return sum(buffer.count(b'\n') for buffer in buffer_gen)


def file_progress_gen(main_window, file: IO, title: str, label: str,
                      no_lines: int, step: int = 100) \
                      -> Generator[Tuple[int, str], None, None]:
    """A generator for file iteration that displays a progress bar

    Yields (line number, line string).
    Return value on user cancel via progress dialog is current line number

    :param main_window: Application main window
    :param file: File to be iterater over
    :param title: Progress dialog title
    :param label: Progress dialog label
    :param no_lines: Number of lines that are remaining in file
    :param step: Number of lines per progress bar update

    """

    with progress_dialog(main_window, title, label, no_lines) as progress_dlg:
        try:
            for i, line in enumerate(file):
                yield i, line

                if not i % step:
                    progress_dlg.setValue(i)
                    QApplication.instance().processEvents()
                    if progress_dlg.wasCanceled():
                        msg = "File operation canceled at line {}.".format(i)
                        raise ProgressDialogCanceled(msg)
        except GeneratorExit:  # Catch for cleaning up with progress_dialog
            pass
        except Exception as error:
            main_window.grid.model.reset()
            main_window.statusBar().showMessage(str(error))



# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread. If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

"""
charts
======

Provides matplotlib figure that are chart templates

Provides
--------

* ChartFigure: Main chart class

"""

from collections import OrderedDict
from copy import copy
from io import StringIO
import datetime
from pathlib import Path
from typing import IO, List, Union

try:
    from matplotlib.figure import Figure
    from matplotlib.sankey import Sankey
    from matplotlib import dates
except ImportError:
    Figure = Sankey = dates = object


def fig2x(figure: Figure, format: Union[str, Path, IO]) -> str:
    """Returns svg from matplotlib chart

    :param figure: Matplotlib figure object
    :param format: matplotlib.pyplot.savefig format, normally filename suffix

    """

    # Save svg to file like object svg_io
    io = StringIO()
    figure.savefig(io, format=format)

    # Rewind the file like object
    io.seek(0)

    data = io.getvalue()
    io.close()

    return data


class ChartFigure(Figure):
    """Chart figure class with drawing method

    **This class is deprecated and exists solely for compatibility with
    pyspread <1.99.0**

    """

    plot_type_fixed_attrs = {
        "plot": ["xdata", "ydata"],
        "bar": ["left", "height"],
        "boxplot": ["x"],
        "hist": ["x"],
        "pie": ["x"],
        "contour": ["X", "Y", "Z"],
        "contourf": ["X", "Y", "Z"],
        "Sankey": [],
    }

    plot_type_xy_mapping = {
        "plot": ["xdata", "ydata"],
        "bar": ["left", "height"],
        "boxplot": ["x", "x"],
        "hist": ["label", "x"],
        "pie": ["labels", "x"],
        "annotate": ["xy", "xy"],
        "contour": ["X", "Y"],
        "contourf": ["X", "Y", "Z"],
        "Sankey": ["flows", "orientations"],
    }

    contour_label_attrs = {
        "contour_labels": "contour_labels",
        "contour_label_fontsize": "fontsize",
        "contour_label_colors": "colors",
    }

    contourf_attrs = {
        "contour_fill": "contour_fill",
        "hatches": "hatches",
    }

    def __init__(self, *attributes: List[dict]):
        """
        :param attributes: List of dicts that contain matplotlib attributes
                           The first list element is defining the axes
                           The following list elements are defining plots

        """

        Figure.__init__(self, (5.0, 4.0), facecolor="white")

        self.attributes = attributes

        self.__axes = self.add_subplot(111)

        # Insert empty attributes with a dict for figure attributes
        if not self.attributes:
            self.attributes = [{}]

        self.draw_chart()

        self.tight_layout(pad=1.5)

    def _xdate_setter(self, xdate_format: str = '%Y-%m-%d'):
        """Makes x axis a date axis with auto format

        :param xdate_format: Sets date formatting

        """

        if xdate_format:
            # We have to validate xdate_format. If wrong then bail out.
            try:
                self.autofmt_xdate()
                datetime.date(2000, 1, 1).strftime(xdate_format)

            except ValueError:
                self.autofmt_xdate()
                return

            self.__axes.xaxis_date()
            formatter = dates.DateFormatter(xdate_format)
            self.__axes.xaxis.set_major_formatter(formatter)

            # The autofmt method does not work in matplotlib 1.3.0
            # self.autofmt_xdate()

    def _setup_axes(self, axes_data: dict):
        """Sets up axes for drawing chart

        :param axes_data: Dicts with keys that match matplotlib axes attributes

        """

        self.__axes.clear()

        key_setter = [
            ("title", self.__axes.set_title),
            ("xlabel", self.__axes.set_xlabel),
            ("ylabel", self.__axes.set_ylabel),
            ("xscale", self.__axes.set_xscale),
            ("yscale", self.__axes.set_yscale),
            ("xticks", self.__axes.set_xticks),
            ("xtick_labels", self.__axes.set_xticklabels),
            ("xtick_params", self.__axes.tick_params),
            ("yticks", self.__axes.set_yticks),
            ("ytick_labels", self.__axes.set_yticklabels),
            ("ytick_params", self.__axes.tick_params),
            ("xlim", self.__axes.set_xlim),
            ("ylim", self.__axes.set_ylim),
            ("xgrid", self.__axes.xaxis.grid),
            ("ygrid", self.__axes.yaxis.grid),
            ("xdate_format", self._xdate_setter),
        ]

        key2setter = OrderedDict(key_setter)

        for key in key2setter:
            if key in axes_data and axes_data[key]:
                try:
                    kwargs_key = key + "_kwargs"
                    kwargs = axes_data[kwargs_key]

                except KeyError:
                    kwargs = {}

                if key == "title":
                    # Shift title up
                    kwargs["y"] = 1.08

                key2setter[key](axes_data[key], **kwargs)

    def _setup_legend(self, axes_data: dict):
        """Sets up legend for drawing chart

        :param axes_data: Dicts with keys that match matplotlib axes attributes

        """

        if "legend" in axes_data and axes_data["legend"]:
            self.__axes.legend()

    def draw_chart(self):
        """Plots chart from self.attributes"""

        if not hasattr(self, "attributes"):
            return

        # The first element is always axes data
        self._setup_axes(self.attributes[0])

        for attribute in self.attributes[1:]:

            series = copy(attribute)
            # Extract chart type
            chart_type_string = series.pop("type")

            x_str, y_str = self.plot_type_xy_mapping[chart_type_string]
            # Check xdata length
            if x_str in series and \
               len(series[x_str]) != len(series[y_str]):
                # Wrong length --> ignore xdata
                series[x_str] = list(range(len(series[y_str])))
            else:
                # Solve the problem that the series data may contain utf-8 data
                series_list = list(series[x_str])
                series_unicode_list = []
                for ele in series_list:
                    if isinstance(ele, bytes):
                        try:
                            series_unicode_list.append(ele.decode('utf-8'))
                        except Exception:
                            series_unicode_list.append(ele)
                    else:
                        series_unicode_list.append(ele)
                series[x_str] = tuple(series_unicode_list)

            fixed_attrs = []
            if chart_type_string in self.plot_type_fixed_attrs:
                for attr in self.plot_type_fixed_attrs[chart_type_string]:
                    # Remove attr if it is a fixed (non-kwd) attr
                    # If a fixed attr is missing, insert a dummy
                    try:
                        fixed_attrs.append(tuple(series.pop(attr)))
                    except KeyError:
                        fixed_attrs.append(())

            # Remove contour chart label info from series
            cl_attrs = {}
            for contour_label_attr in self.contour_label_attrs:
                if contour_label_attr in series:
                    cl_attrs[self.contour_label_attrs[contour_label_attr]] = \
                        series.pop(contour_label_attr)

            # Remove contourf attributes from series
            cf_attrs = {}
            for contourf_attr in self.contourf_attrs:
                if contourf_attr in series:
                    cf_attrs[self.contourf_attrs[contourf_attr]] = \
                        series.pop(contourf_attr)

            if not fixed_attrs or all(fixed_attrs):
                # Draw series to axes

                # Do we have a Sankey plot --> build it
                if chart_type_string == "Sankey":
                    Sankey(self.__axes, **series).finish()

                else:
                    chart_method = getattr(self.__axes, chart_type_string)
                    plot = chart_method(*fixed_attrs, **series)

                # Do we have a filled contour?
                try:
                    if cf_attrs.pop("contour_fill"):
                        cf_attrs.update(series)
                        if "linewidths" in cf_attrs:
                            cf_attrs.pop("linewidths")
                        if "linestyles" in cf_attrs:
                            cf_attrs.pop("linestyles")
                        if not cf_attrs["hatches"]:
                            cf_attrs.pop("hatches")
                        self.__axes.contourf(plot, **cf_attrs)
                except KeyError:
                    pass

                # Do we have a contour chart label?
                try:
                    if cl_attrs.pop("contour_labels"):
                        self.__axes.clabel(plot, **cl_attrs)
                except KeyError:
                    pass

        # The legend has to be set up after all series are drawn
        self._setup_legend(self.attributes[0])



from rpy2 import robjects
from rpy2.interactive import process_revents
from rpy2.robjects import Formula
from rpy2.robjects.packages import importr, data
import rpy2.robjects.lib.ggplot2 as ggplot2
from rpy2.robjects.lib import grdevices

try:
    process_revents.start()
except:
    pass

base = importr('base')
datasets = importr('datasets')

mtcars = data(datasets).fetch('mtcars')['mtcars']

plot = ggplot2.ggplot(mtcars)
plot += ggplot2.aes_string(x='factor(cyl)', y='mpg', col='factor(cyl)')
plot += ggplot2.geom_boxplot()
plot += ggplot2.ggtitle('Box plots')

with grdevices.render_to_bytesio(grdevices.svg) as svg:
    plot.plot()
process_revents.stop()

svg.getvalue() 



# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

"""

**Provides**

* :class:`Module`
* :class:`DependenciesDialog`
* :class:`InstallPackageDialog`

"""


try:
    from dataclasses import dataclass
except ImportError:
    # Python 3.6 compatibility
    from pyspread.lib.dataclasses import dataclass
import os

from importlib.metadata import version as version__
from importlib.metadata import PackageNotFoundError

from PyQt6.QtCore import QProcess, QSize
from PyQt6.QtGui import QFont, QTextCursor
from PyQt6.QtWidgets import (
        QDialog, QButtonGroup, QVBoxLayout, QHBoxLayout, QTreeWidgetItem,
        QToolButton, QGroupBox, QTreeWidget, QCheckBox, QLineEdit,
        QPlainTextEdit, QWidget, QPushButton, QAbstractScrollArea)

try:
    from packaging import version
except ImportError:
    # We fall back to local library to remove the dependency
    try:
        from pyspread.lib.packaging import version
    except ImportError:
        from lib.packaging import version
try:
    from pyspread.lib.attrdict import AttrDict
except ImportError:
    from lib.attrdict import AttrDict


@dataclass
class Module:
    """Module checker"""

    name: str
    description: str
    required_version: str  # The minimum version number that is required

    @property
    def version(self) -> version:
        """Currently installed version number, False if not installed"""

        try:
            return version.parse(version__(self.name))
        except PackageNotFoundError:
            return

    def is_installed(self) -> bool:
        """True if the module is installed"""

        __version = self.version

        return bool(__version) if __version is not None else None

# Required dependencies
# ---------------------

# Required dependencies are checked by the cli


REQUIRED_DEPENDENCIES = [
    Module(name="markdown2",
           description="Python implementation of Markdown.",
           required_version=version.parse("2.3")),
    Module(name="numpy",
           description="Fundamental package for scientific computing.",
           required_version=version.parse("1.1")),
    Module(name="PyQt6",
           description="Python bindings for the Qt application framework.",
           required_version=version.parse("6.5")),
    Module(name="setuptools",
           description="Download, build, install, upgrade, and uninstall "
           "Python packages.",
           required_version=version.parse("40.0")),
]

# Optional dependencies
# ---------------------


OPTIONAL_DEPENDENCIES = [
    Module(name="matplotlib",
           description="Create visualizations in Python.",
           required_version=version.parse("3.4")),
    Module(name="openpyxl",
           description="Read Excel 2010 xlsx files.",
           required_version=version.parse("3.0.7")),
    Module(name="plotnine",
           description="Grammar of graphics for simpler R ggplot2 charts.",
           required_version=version.parse("0.8")),
    Module(name="pycel",
           description="Translate an Excel spreadsheet into executable python "
                       "code.",
           required_version=version.parse("1.0b30")),
    Module(name="python-dateutil",
           description="Parse date strings.",
           required_version=version.parse("2.7.0")),
    Module(name="pyenchant",
           description="Bindings for the Enchant spellchecking library.",
           required_version=version.parse("1.1")),
    Module(name="py-moneyed",
           description="Money class for csv import and usage in cells.",
           required_version=version.parse("2.0")),
    Module(name="rpy2",
           description="Interface to R, required for R charts.",
           required_version=version.parse("3.4")),
]

DEPENDENCIES = REQUIRED_DEPENDENCIES + OPTIONAL_DEPENDENCIES

PIP_MODULE = Module(name="pip", description="pip installer",
                    required_version=version.parse("17.0"))

# Not yet implemented modules
#    Module(name="xlrd",
#           description="Load Excel files",
#           required_version="0.9.2"),
#    Module(name="xlwt",
#           description="Save Excel files",
#           required_version="0.9.2"),


class DependenciesDialog(QDialog):
    """Dependencies dialog for python dependencies"""

    column = AttrDict(zip(("button", "status", "name", "version",
                           "required_version", "description"), range(6)))
    column_headers = ("", "Status", "Package", "Version", "Required",
                      "Description")

    def __init__(self, parent: QWidget = None):
        """
        :param parent: Parent widget

        """

        super().__init__(parent)

        self.setWindowTitle("Installer")

        # Button group for install buttons
        self.buttGroup = QButtonGroup()
        self.buttGroup.buttonClicked.connect(self.on_butt_install)

        self.mainLayout = QVBoxLayout()
        self.mainLayout.setContentsMargins(10, 10, 10, 10)
        self.setLayout(self.mainLayout)

        self.tree = QTreeWidget()
        self.tree.setSizeAdjustPolicy(QAbstractScrollArea.SizeAdjustPolicy.AdjustToContents)
        self.mainLayout.addWidget(self.tree, 4)

        self.tree.setHeaderLabels(self.column_headers)
        self.tree.setRootIsDecorated(False)
        self.tree.setSelectionMode(QTreeWidget.SelectionMode.NoSelection)

        self.update_load()

    def update_load(self):

        self.tree.clear()

        for module in DEPENDENCIES:
            item = QTreeWidgetItem()
            if module in REQUIRED_DEPENDENCIES:
                item.setText(self.column.button, "Required")
            else:
                item.setText(self.column.button, "Optional")
            item.setText(self.column.name, module.name)
            version = module.version if module.version else "Not installed"
            item.setText(self.column.version, str(version))
            item.setText(self.column.required_version,
                         str(module.required_version))
            item.setText(self.column.description, module.description)
            self.tree.addTopLevelItem(item)

            if module.is_installed():
                status = "Installed"
                bold = False
            elif module.is_installed() is None:
                status = "pkg_resources is missing"
                bold = True
            else:
                status = "Not installed"
                bold = True
                butt = QToolButton()
                butt.setText("Install")
                butt.setEnabled(PIP_MODULE.is_installed())
                self.tree.setItemWidget(item, self.column.button, butt)
                self.buttGroup.addButton(butt, DEPENDENCIES.index(module))

            item.setText(self.column.status, status)
            if bold:
                item.setFont(self.column.status, QFont("", -1, QFont.Weight.Bold))

        for col in range(self.tree.columnCount()):
            self.tree.resizeColumnToContents(col)

    def on_butt_install(self, butt: QPushButton):
        """One of install buttons pressed

        :param butt: The pressed button

        """

        butt.setDisabled(True)
        idx = self.buttGroup.id(butt)

        dial = InstallPackageDialog(self, module=DEPENDENCIES[idx])
        dial.exec()
        self.update_load()


class InstallPackageDialog(QDialog):
    """Shows a dialog to execute command"""

    line_str = "-" * 56

    def __init__(self, parent=None, module=None):
        """
        :param parent: Parent widget
        :param module: Module to be installed

        """

        super().__init__(parent)

        self.module = module

        self.setWindowTitle("Install Package")
        self.setMinimumWidth(600)

        self.process = QProcess(self)
        self.process.readyReadStandardOutput.connect(self.on_read_standard)
        self.process.readyReadStandardError.connect(self.on_read_error)
        self.process.finished.connect(self.on_finished)

        self.mainLayout = QVBoxLayout()
        self.mainLayout.setContentsMargins(10, 10, 10, 10)
        self.setLayout(self.mainLayout)

        self.groupBox = QGroupBox()
        self.groupBox.setTitle("Shell Command")
        self.groupBoxLayout = QHBoxLayout()
        self.groupBox.setLayout(self.groupBoxLayout)
        self.mainLayout.addWidget(self.groupBox)

        self.buttSudo = QCheckBox()
        self.buttSudo.setText("sudo")
        self.groupBoxLayout.addWidget(self.buttSudo, 0)
        self.buttSudo.toggled.connect(self.update_cmd_line)
        self.buttSudo.setVisible(os.name != "nt")

        self.txtCommand = QLineEdit()
        self.groupBoxLayout.addWidget(self.txtCommand, 10)

        self.buttExecute = QPushButton()
        self.buttExecute.setText("Execute")
        self.groupBoxLayout.addWidget(self.buttExecute, 0)
        self.buttExecute.clicked.connect(self.on_butt_execute)

        self.txtStdOut = QPlainTextEdit()
        self.mainLayout.addWidget(self.txtStdOut)

        self.txtStdErr = QPlainTextEdit()
        self.mainLayout.addWidget(self.txtStdErr)

        self.update_cmd_line()

    def update_cmd_line(self, *unused):
        """Update the commend line considering sudo button state"""

        cmd = ""
        if self.buttSudo.isChecked():
            cmd += "pkexec  "

        cmd += "pip3 install {modulename}".format(modulename=self.module.name)

        self.txtCommand.setText(cmd)

    def on_butt_execute(self):
        """Execute button event handler"""

        self.buttSudo.setDisabled(True)
        self.buttExecute.setDisabled(True)

        self.txtStdOut.setPlainText("")
        self.txtStdErr.setPlainText("")
        self.process.start(self.txtCommand.text())

    def on_read_standard(self):
        """Stdout read event handler"""

        msg_tpl = "{}\n{}\n{}"
        msg = msg_tpl.format(self.txtStdOut.toPlainText(),
                             self.line_str,
                             self.process.readAllStandardOutput())

        self.txtStdOut.setPlainText(msg)
        self.txtStdOut.moveCursor(QTextCursor.End)

    def on_read_error(self):
        """Stderr read event handler"""

        msg_tpl = "{}\n{}\n{}"
        msg = msg_tpl.format(self.txtStdErr.toPlainText(),
                             self.line_str,
                             self.process.readAllStandardError())

        self.txtStdErr.setPlainText(msg)
        self.txtStdErr.moveCursor(QTextCursor.End)

    def on_finished(self):
        """Execution finished event handler"""

        self.buttSudo.setDisabled(False)
        self.buttExecute.setDisabled(False)



# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

"""

**Provides**

 * :func: `painter_save`: Context manager saving and restoring painter state
 * :func: `painter_zoom`: Context manager scaling and restoring the painter
 * :func: `painter_rotate`: Context manager rotating and restoring the painter
 * :class:`GridCellNavigator`: Find neighbors of a cell
 * :class:`EdgeBorders`: Dataclass for edge properties
 * :class:`CellEdgeRenderer`: Paints cell edges
 * :class:`QColorCache`: QColor cache
 * :class:`CellRenderer`: Paints cells

"""

from contextlib import contextmanager
try:
    from dataclasses import dataclass
except ImportError:
    from pyspread.lib.dataclasses import dataclass  # Python 3.6 compatibility
from typing import List, Tuple

try:
    from pyspread.themes import ColorRole
except ImportError:
    from themes import ColorRole

from PyQt6.QtCore import Qt, QModelIndex, QRectF, QPointF
from PyQt6.QtGui import (QBrush, QPainter, QPalette, QPen,
                         QPainterPath, QPolygonF, QPainterPathStroker)

from PyQt6.QtGui import QColor as __QColor
from PyQt6.QtWidgets import QTableView, QStyleOptionViewItem
from functools import lru_cache


@contextmanager
def painter_save(painter: QPainter):
    """Context manager saving and restoring painter state

    :param painter: Painter, for which the state is preserved

    """

    painter.save()
    yield
    painter.restore()


@contextmanager
def painter_zoom(painter: QPainter, zoom: float, rect: QRectF) -> QRectF:
    """Context manager scaling and restoring the painter

    (rect.x(), rect.y()) is invariant

    :param painter: Painter, for which the state is preserved
    :param zoom: Zoom factor
    :param rect: Rect for setting zoom invariant point (rect.x(), rect.y())

    """

    with painter_save(painter):
        painter.translate(rect.x(), rect.y())
        painter.scale(zoom, zoom)
        painter.translate(-rect.x() * zoom, -rect.y() * zoom)
        yield QRectF(rect.x() * zoom, rect.y() * zoom,
                     rect.width() / zoom, rect.height() / zoom)


@contextmanager
def painter_rotate(painter: QPainter, rect: QRectF, angle: int = 0) -> QRectF:
    """Context manager rotating and restoring the painter

    :param painter: Painter, which is rotated
    :param rect: Rect to be painted in
    :param angle: Rotataion angle must be in (0, 90, 180, 270)

    """

    supported_angles = 0, 90, 180, 270
    angle = int(angle)

    if angle not in supported_angles:
        msg = "Rotation angle {} not in {}".format(angle, supported_angles)
        raise Warning(msg)
        return

    center_x, center_y = rect.center().x(), rect.center().y()

    with painter_save(painter):
        painter.translate(center_x, center_y)
        painter.rotate(angle)

        if angle in (0, 180):
            painter.translate(-center_x, -center_y)
        elif angle in (90, 270):
            painter.translate(-center_y, -center_x)
            rect = QRectF(rect.y(), rect.x(), rect.height(), rect.width())
        yield rect


class QColor(__QColor):
    """Hashable QColor"""

    def __hash__(self):
        return self.rgba()


class GridCellNavigator:
    """Find neighbors of a cell"""

    def __init__(self, grid: QTableView, key: Tuple[int, int, int]):
        """
        :param grid: The main grid widget
        :param key: Key of cell fow which neighbors are identified

        """

        self.grid = grid
        self.code_array = grid.model.code_array
        self.row, self.column, self.table = self.key = key

        self.borderwidth_bottom_cache = grid.borderwidth_bottom_cache
        self.borderwidth_right_cache = grid.borderwidth_right_cache

    @property
    def borderwidth_bottom(self) -> float:
        """Width of bottom border line"""

        return self.borderwidth_bottom_cache[self.key]

    @property
    def borderwidth_right(self) -> float:
        """Width of right border line"""

        return self.borderwidth_right_cache[self.key]

    @property
    def border_color_bottom(self) -> QColor:
        """Color of bottom border line"""

        return self.code_array.cell_attributes[self.key].bordercolor_bottom

    @property
    def border_color_right(self) -> QColor:
        """Color of right border line"""

        return self.code_array.cell_attributes[self.key].bordercolor_right

    @property
    def merge_area(self) -> Tuple[int, int, int, int]:
        """Merge area of the key cell"""

        return self.code_array.cell_attributes[self.key].merge_area

    def merging_key(self, key: Tuple[int, int, int]) -> Tuple[int, int, int]:
        """Merging cell if cell is merged else cell key

        :param key: Key of cell that is checked for being merged

        """

        merging_key = self.code_array.cell_attributes.get_merging_cell(key)
        return key if merging_key is None else merging_key

    def above_keys(self) -> List[Tuple[int, int, int]]:
        """Key list of neighboring cells above the key cell"""

        merge_area = self.merge_area
        if merge_area is None:
            return [self.merging_key((self.row - 1, self.column, self.table))]
        _, left, _, right = merge_area
        return [self.merging_key((self.row - 1, col, self.table))
                for col in range(left, right + 1)]

    def below_keys(self) -> List[Tuple[int, int, int]]:
        """Key list of neighboring cells below the key cell"""

        merge_area = self.merge_area
        if merge_area is None:
            return [self.merging_key((self.row + 1, self.column, self.table))]
        _, left, _, right = merge_area
        return [self.merging_key((self.row + 1, col, self.table))
                for col in range(left, right + 1)]

    def left_keys(self) -> List[Tuple[int, int, int]]:
        """Key list of neighboring cells left of the key cell"""

        merge_area = self.merge_area
        if merge_area is None:
            return [self.merging_key((self.row, self.column - 1, self.table))]
        top, _, bottom, _ = merge_area
        return [self.merging_key((row, self.column - 1, self.table))
                for row in range(top, bottom + 1)]

    def right_keys(self) -> List[Tuple[int, int, int]]:
        """Key list of neighboring cells right of the key cell"""

        merge_area = self.merge_area
        if merge_area is None:
            return [self.merging_key((self.row, self.column + 1, self.table))]
        top, _, bottom, _ = merge_area
        return [self.merging_key((row, self.column + 1, self.table))
                for row in range(top, bottom + 1)]

    def above_left_key(self) -> Tuple[int, int, int]:
        """Key of neighboring cell above left of the key cell"""

        return self.merging_key((self.row - 1, self.column - 1, self.table))

    def above_right_key(self) -> Tuple[int, int, int]:
        """Key of neighboring cell above right of the key cell"""

        return self.merging_key((self.row - 1, self.column + 1, self.table))

    def below_left_key(self) -> Tuple[int, int, int]:
        """Key of neighboring cell below left of the key cell"""

        return self.merging_key((self.row + 1, self.column - 1, self.table))

    def below_right_key(self) -> Tuple[int, int, int]:
        """Key of neighboring cell below right of the key cell"""

        return self.merging_key((self.row + 1, self.column + 1, self.table))


@dataclass
class EdgeBorders:
    """Holds border data for an edge and provides effective edge properties"""

    left_width: float
    right_width: float
    top_width: float
    bottom_width: float

    left_color: QColor
    right_color: QColor
    top_color: QColor
    bottom_color: QColor

    left_x: float
    right_x: float
    top_y: float
    bottom_y: float

    _color_cache = None

    @property
    def _border_widths(self) -> Tuple[float, float, float, float]:
        """Tuple of border widths in order left, right, top, bottom"""

        return (self.left_width, self.right_width,
                self.top_width, self.bottom_width)

    @property
    def _border_colors(self) -> Tuple[QColor, QColor, QColor, QColor]:
        """Tuple of border colors in order left, right, top, bottom"""

        return (self.left_color, self.right_color,
                self.top_color, self.bottom_color)

    @property
    def width(self) -> float:
        """Edge width, i.e. thickest vertical border"""

        return max(self.top_width, self.bottom_width)

    @property
    def height(self) -> float:
        """Edge height, i.e. thickest horizontal border"""

        return max(self.left_width, self.right_width)

    @property
    def color(self) -> QColor:
        """Edge color, i.e. darkest color of thickest edge border"""

        if self._color_cache is not None:
            return self._color_cache

        max_border_width = max(self.width, self.height)
        colors = []
        for width, color in zip(self._border_widths, self._border_colors):
            if width == max_border_width:
                colors.append(color)
        colors.sort(key=lambda color: color.lightnessF())
        color = colors[0]

        self._color_cache = color

        return color


class CellEdgeRenderer:
    """Paints cell edges"""

    intersection_cache = {}

    def __init__(self, painter: QPainter, center: QPointF,
                 borders: EdgeBorders, rect: QRectF, clip_path: QPainterPath,
                 zoom: float):
        """

        Borders are provided by EdgeBorders in order: left, right, top, bottom

        :param painter: Painter with which edge is drawn
        :param center: Edge center
        :param borders: Border widths and colors
        :param rect: Rect of the clip_path
        :param clip_path: Clip rectangle that is required for QtSVG clipping
        :param zoom: Current zoom level

        """

        self.painter = painter
        self.center = center
        self.borders = borders
        self.rect = rect
        self.clip_path = clip_path
        self.zoom = zoom

    def paint(self):
        """Paints the edge"""

        if not self.borders.width or not self.borders.height:
            return  # Invisible edge

        x, y = self.center.x(), self.center.y()
        width = self.borders.width * self.zoom
        height = self.borders.height * self.zoom

        cache_key = (self.rect.x(), self.rect.y(),
                     self.rect.width(), self.rect.height(),
                     x, y, width, height)

        color = self.borders.color
        self.painter.setPen(QPen(Qt.PenStyle.NoPen))
        self.painter.setBrush(QBrush(color))

        try:
            intersection = self.intersection_cache[cache_key]
        except KeyError:
            rect = QRectF(x-width/2, y-height/2, width, height)
            rect_path = QPainterPath()  # Required for clipping in SVG export
            rect_path.addRect(rect)

            intersection = self.clip_path.intersected(rect_path)

            self.intersection_cache[cache_key] = intersection

        self.painter.drawPath(intersection)
        self.painter.setPen(QPen(Qt.PenStyle.SolidLine))


class QColorCache(dict):
    """QColor cache that returns placeholder text color for None"""

    def __init__(self, grid, *args, **kwargs):
        self.grid = grid
        super().__init__(*args, **kwargs)

    def __missing__(self, key):
        if key is None:
            qcolor = QColor(self.grid.palette().color(ColorRole.line))
        else:
            qcolor = QColor(*key)

        self[key] = qcolor

        return qcolor


class BorderWidthBottomCache(dict):
    """BorderWidthBottom cache"""

    def __init__(self, grid, *args, **kwargs):
        self.grid = grid
        self.cell_attributes = grid.model.code_array.cell_attributes

        super().__init__(*args, **kwargs)

    def __missing__(self, key):
        borderwidth_bottom = self.cell_attributes[key].borderwidth_bottom
        self[key] = borderwidth_bottom

        return borderwidth_bottom


class BorderWidthRightCache(BorderWidthBottomCache):
    """BorderWidthRight cache"""

    def __missing__(self, key):
        borderwidth_right = self.cell_attributes[key].borderwidth_right
        self[key] = borderwidth_right

        return borderwidth_right


class EdgeBordersCache(dict):
    """Cache of all EdgeBorders objects"""

    def __missing__(self, key):
        self[key] = edge_border = EdgeBorders(*key)

        return edge_border


class BorderColorBottomCache(BorderWidthBottomCache):
    """BorderColorBottomCache cache"""

    def __missing__(self, key):
        border_color_bottom = self.cell_attributes[key].border_color_bottom
        self[key] = border_color_bottom

        return border_color_bottom


class BorderColorRightCache(BorderWidthBottomCache):
    """BorderColorBottomCache cache"""

    def __missing__(self, key):
        border_color_right = self.cell_attributes[key].border_color_right
        self[key] = border_color_right

        return border_color_right


class CellRenderer:
    """Paints cells

    Cell rendering governs the area of a cell inside its borders.
    It is done in a  vector oriented way.
    Therefore, the following conventions shall apply:
     * Cell borders of width 1 shall be painted so that they appear only in the
       bottom and right edge of the cell.
     * Cell borders of all widths have the same center line.
     * Cell borders that are thicker than 1 are painted on all borders.
     * At the edges, borders are painted in the following order:
        1. Thin borders are painted first
        2. If border width is equal, lighter colors are painted first

    """

    def __init__(self, grid: QTableView, painter: QPainter,
                 option: QStyleOptionViewItem, index: QModelIndex):
        """
        :param grid: The main grid widget
        :param painter: Painter with which borders are drawn
        :param option: Style option for rendering
        :param index: Index of cell to be rendered
        """

        self.grid = grid
        self.painter = painter
        self.option = option
        self.index = index

        self.painter.setRenderHint(QPainter.RenderHint.LosslessImageRendering,
                                   True)
        self.painter.setRenderHint(QPainter.RenderHint.SmoothPixmapTransform,
                                   True)

        self.cell_attributes = grid.model.code_array.cell_attributes
        self.key = index.row(), index.column(), self.grid.table

        self.cell_nav = GridCellNavigator(grid, self.key)

        self.qcolor_cache = grid.qcolor_cache
        self.borderwidth_bottom_cache = grid.borderwidth_bottom_cache
        self.borderwidth_right_cache = grid.borderwidth_right_cache

    def inner_rect(self, rect: QRectF) -> QRectF:
        """Returns inner rect that is shrunk by border widths

        For merged cells, minimum top/left border widths are taken into account

        :param rect: Cell rect to be shrunk

        """

        above_keys = self.cell_nav.above_keys()
        left_keys = self.cell_nav.left_keys()

        width_top = min(self.borderwidth_bottom_cache[above_key]
                        for above_key in above_keys)
        width_left = min(self.borderwidth_right_cache[left_key]
                         for left_key in left_keys)
        width_bottom = self.cell_nav.borderwidth_bottom
        width_right = self.cell_nav.borderwidth_right

        width_top *= self.grid.zoom
        width_left *= self.grid.zoom
        width_bottom *= self.grid.zoom
        width_right *= self.grid.zoom

        rect_x = rect.x() + width_left / 2
        rect_y = rect.y() + width_top / 2
        rect_width = rect.width() - width_left / 2 - width_right / 2
        rect_height = rect.height() - width_top / 2 - width_bottom / 2

        return QRectF(rect_x, rect_y, rect_width, rect_height)

    def paint_content(self, rect: QRectF):
        """
        :param rect: Cell rect of the cell to be painted

        """

        with painter_zoom(self.painter, self.grid.zoom, rect) as zrect:
            self.grid.delegate.paint_(self.painter, zrect, self.option,
                                      self.index)

    @staticmethod
    @lru_cache(maxsize=65536)
    def _get_border_pen(width, zoom):
        """Gets zoomed border pen, white, fully transparent

        :param width: Unzoomed line width
        :param zoom: Current zoom level of grid

        """

        return QPen(QColor(255, 255, 255, 0), zoom*width,
                    Qt.PenStyle.SolidLine, Qt.PenCapStyle.FlatCap,
                    Qt.PenJoinStyle.MiterJoin)

    def _draw_line(self, point1: QPointF, point2: QPointF, width: float,
                   color: QColor, clip_path: QPainterPath):
        """Draws line

        Uses drawLine for screen painting and QPainterPathStroker for
        svg painting.

        :param point1: Start point of line
        :param point2: End point of line
        :param width: Line width
        :param color: Line color
        :param clip_path: Clip rectangle

        """

        zoom = self.grid.zoom
        alpha = max(0, round(255 - 255 * width * zoom))

        if width * zoom > 1.01:
            self.painter.setRenderHint(QPainter.RenderHint.Antialiasing, True)

        screen_painting = self.grid.main_window.print_area is None

        if screen_painting:
            # Rendering for the screen
            pen = QPen(color)
            pen.setWidthF(width * zoom)
            self.painter.setPen(pen)
            self.painter.drawLine(point1, point2)
        else:
            # Rendering for the printer
            pen = self._get_border_pen(width, zoom)

            line_polygon = QPolygonF((point1, point2))
            line_path = QPainterPath()
            line_path.addPolygon(line_polygon)

            stroker = QPainterPathStroker(pen)
            stroked_path = stroker.createStroke(line_path)

            self.painter.setPen(QPen(QColor(255, 255, 255, alpha)))
            self.painter.setBrush(QBrush(color))
            self.painter.drawPath(clip_path.intersected(stroked_path))

        self.painter.setRenderHint(QPainter.RenderHint.Antialiasing, False)

    def paint_bottom_border(self, rect: QRectF, clip_path: QPainterPath):
        """Paint bottom border of cell

        :param rect: Cell rect of the cell to be painted
        :param clip_path: Clip rectangle that is required for QtSVG clipping

        """

        if not self.cell_nav.borderwidth_bottom:
            return

        point1 = QPointF(rect.x(), rect.y() + rect.height())
        point2 = QPointF(rect.x() + rect.width(), rect.y() + rect.height())

        width = self.cell_nav.borderwidth_bottom
        color = self.qcolor_cache[self.cell_nav.border_color_bottom]

        self._draw_line(point1, point2, width, color, clip_path)

    def paint_right_border(self, rect: QRectF, clip_path: QPainterPath):
        """Paint right border of cell

        :param rect: Cell rect of the cell to be painted
        :param clip_path: Clip rectangle that is requuired for QtSVG clipping

        """

        if not self.cell_nav.borderwidth_right:
            return

        point1 = QPointF(rect.x() + rect.width(), rect.y())
        point2 = QPointF(rect.x() + rect.width(), rect.y() + rect.height())

        width = self.cell_nav.borderwidth_right
        color = self.qcolor_cache[self.cell_nav.border_color_right]

        self._draw_line(point1, point2, width, color, clip_path)

    def paint_above_borders(self, rect: QRectF, clip_path: QPainterPath):
        """Paint lower borders of all above cells

        :param rect: Cell rect of below cell, in which the borders are painted
        :param clip_path: Clip rectangle that is requuired for QtSVG clipping

        """

        for above_key in self.cell_nav.above_keys():
            above_cell_nav = GridCellNavigator(self.grid, above_key)
            if not above_cell_nav.borderwidth_bottom:
                continue

            merge_area = above_cell_nav.merge_area
            if merge_area is None:
                columns = [above_key[1]]
            else:
                _, left, _, right = merge_area
                columns = list(range(left, right + 1))
            above_rect_x = self.grid.columnViewportPosition(columns[0])
            above_rect_width = sum(self.grid.columnWidth(column)
                                   for column in columns)

            point1 = QPointF(above_rect_x, rect.y())
            point2 = QPointF(above_rect_x + above_rect_width, rect.y())

            width = above_cell_nav.borderwidth_bottom
            color = self.qcolor_cache[above_cell_nav.border_color_bottom]

            self._draw_line(point1, point2, width, color, clip_path)

    def paint_left_borders(self, rect: QRectF, clip_path: QPainterPath):
        """Paint right borders of all left cells

        :param rect: Cell rect of right cell, in which the borders are painted
        :param clip_path: Clip rectangle that is requuired for QtSVG clipping

        """

        for left_key in self.cell_nav.left_keys():
            left_cell_nav = GridCellNavigator(self.grid, left_key)
            if not left_cell_nav.borderwidth_right:
                continue

            merge_area = left_cell_nav.merge_area
            if merge_area is None:
                rows = [left_key[0]]
            else:
                top, _, bottom, _ = merge_area
                rows = list(range(top, bottom + 1))
            left_rect_y = self.grid.rowViewportPosition(rows[0])
            left_rect_height = sum(self.grid.rowHeight(row) for row in rows)

            point1 = QPointF(rect.x(), left_rect_y)
            point2 = QPointF(rect.x(), left_rect_y + left_rect_height)

            width = left_cell_nav.borderwidth_right
            color = self.qcolor_cache[left_cell_nav.border_color_right]

            self._draw_line(point1, point2, width, color, clip_path)

    def paint_top_left_edge(self, rect: QRectF, clip_path: QPainterPath):
        """Paints top left edge of the cell

        :param rect: Cell rect of cell, for which the edge is painted
        :param clip_path: Clip rectangle that is requuired for QtSVG clipping

                  top
               TL  |  T
        left ------------ right
               L   |  C
                 bottom

        """

        center = QPointF(rect.x(), rect.y())

        top_left_key = self.cell_nav.above_left_key()
        left_key = self.cell_nav.left_keys()[0]
        top_key = self.cell_nav.above_keys()[0]

        top_left_cell_nav = GridCellNavigator(self.grid, top_left_key)
        left_cell_nav = GridCellNavigator(self.grid, left_key)
        top_cell_nav = GridCellNavigator(self.grid, top_key)

        left_width = top_left_cell_nav.borderwidth_bottom
        right_width = top_cell_nav.borderwidth_bottom
        top_width = top_left_cell_nav.borderwidth_right
        bottom_width = left_cell_nav.borderwidth_right

        left_color = self.qcolor_cache[top_left_cell_nav.border_color_bottom]
        right_color = self.qcolor_cache[top_cell_nav.border_color_bottom]
        top_color = self.qcolor_cache[top_left_cell_nav.border_color_right]
        bottom_color = self.qcolor_cache[left_cell_nav.border_color_right]

        left_x = rect.x() - rect.width()
        right_x = rect.x()
        top_y = rect.y() - rect.height()
        bottom_y = rect.y()

        key = (left_width, right_width, top_width, bottom_width,
               left_color, right_color, top_color, bottom_color,
               left_x, right_x, top_y, bottom_y)

        borders = self.grid.edge_borders_cache[key]

        renderer = CellEdgeRenderer(self.painter, center, borders, rect,
                                    clip_path, self.grid.zoom)
        renderer.paint()

    def paint_top_right_edge(self, rect: QRectF, clip_path: QPainterPath):
        """Paints top right edge of the cell

        :param rect: Cell rect of cell, for which the edge is painted
        :param clip_path: Clip rectangle that is requuired for QtSVG clipping

                  top
                T  |  TR
        left ------------ right
                C  |  R
                 bottom

        """

        center = QPointF(rect.x() + rect.width(), rect.y())

        top_key = self.cell_nav.above_keys()[-1]
        top_right_key = self.cell_nav.above_right_key()

        top_cell_nav = GridCellNavigator(self.grid, top_key)
        top_right_cell_nav = GridCellNavigator(self.grid, top_right_key)

        left_width = top_cell_nav.borderwidth_bottom
        right_width = top_right_cell_nav.borderwidth_bottom
        top_width = top_cell_nav.borderwidth_right
        bottom_width = self.cell_nav.borderwidth_right

        left_color = self.qcolor_cache[top_cell_nav.border_color_bottom]
        right_color = self.qcolor_cache[top_right_cell_nav.border_color_bottom]
        top_color = self.qcolor_cache[top_cell_nav.border_color_right]
        bottom_color = self.qcolor_cache[self.cell_nav.border_color_right]

        left_x = rect.x() + rect.width()
        right_x = rect.x() + 2 * rect.width()
        top_y = rect.y() - rect.height()
        bottom_y = rect.y()

        key = (left_width, right_width, top_width, bottom_width,
               left_color, right_color, top_color, bottom_color,
               left_x, right_x, top_y, bottom_y)

        borders = self.grid.edge_borders_cache[key]

        renderer = CellEdgeRenderer(self.painter, center, borders, rect,
                                    clip_path, self.grid.zoom)
        renderer.paint()

    def paint_bottom_left_edge(self, rect: QRectF, clip_path: QPainterPath):
        """Paints bottom left edge of the cell

        :param rect: Cell rect of cell, for which the edge is painted
        :param clip_path: Clip rectangle that is requuired for QtSVG clipping

                  top
               L   |  C
        left ------------ right
               BL  |  B
                 bottom

        """

        center = QPointF(rect.x(), rect.y() + rect.height())

        left_key = self.cell_nav.left_keys()[-1]
        bottom_left_key = self.cell_nav.below_left_key()

        left_cell_nav = GridCellNavigator(self.grid, left_key)
        bottom_left_cell_nav = GridCellNavigator(self.grid, bottom_left_key)

        left_width = left_cell_nav.borderwidth_bottom
        right_width = self.cell_nav.borderwidth_bottom
        top_width = left_cell_nav.borderwidth_right
        bottom_width = bottom_left_cell_nav.borderwidth_right

        left_color = self.qcolor_cache[left_cell_nav.border_color_bottom]
        right_color = self.qcolor_cache[self.cell_nav.border_color_bottom]
        top_color = self.qcolor_cache[left_cell_nav.border_color_right]
        bottom_color = \
            self.qcolor_cache[bottom_left_cell_nav.border_color_right]

        left_x = rect.x() - rect.width()
        right_x = rect.x()
        top_y = rect.y() + rect.height()
        bottom_y = rect.y() + 2 * rect.height()

        key = (left_width, right_width, top_width, bottom_width,
               left_color, right_color, top_color, bottom_color,
               left_x, right_x, top_y, bottom_y)

        borders = self.grid.edge_borders_cache[key]

        renderer = CellEdgeRenderer(self.painter, center, borders, rect,
                                    clip_path, self.grid.zoom)
        renderer.paint()

    def paint_bottom_right_edge(self, rect: QRectF, clip_path: QPainterPath):
        """Paints bottom right edge of the cell

        :param rect: Cell rect of cell, for which the edge is painted
        :param clip_path: Clip rectangle that is requuired for QtSVG clipping

                 top
               C  |  R
        left ----------- right
               B  |  BR
                bottom

        """

        center = QPointF(rect.x() + rect.width(), rect.y() + rect.height())

        right_key = self.cell_nav.right_keys()[-1]
        bottom_key = self.cell_nav.below_keys()[-1]

        right_cell_nav = GridCellNavigator(self.grid, right_key)
        bottom_cell_nav = GridCellNavigator(self.grid, bottom_key)

        left_width = self.cell_nav.borderwidth_bottom
        right_width = right_cell_nav.borderwidth_bottom
        top_width = self.cell_nav.borderwidth_right
        bottom_width = bottom_cell_nav.borderwidth_right

        left_color = self.qcolor_cache[self.cell_nav.border_color_bottom]
        right_color = self.qcolor_cache[right_cell_nav.border_color_bottom]
        top_color = self.qcolor_cache[self.cell_nav.border_color_right]
        bottom_color = self.qcolor_cache[bottom_cell_nav.border_color_right]

        left_x = rect.x() + rect.width()
        right_x = rect.x() + 2 * rect.width()
        top_y = rect.y() + rect.height()
        bottom_y = rect.y() + 2 * rect.height()

        key = (left_width, right_width, top_width, bottom_width,
               left_color, right_color, top_color, bottom_color,
               left_x, right_x, top_y, bottom_y)

        borders = self.grid.edge_borders_cache[key]

        renderer = CellEdgeRenderer(self.painter, center, borders, rect,
                                    clip_path, self.grid.zoom)
        renderer.paint()

    def paint_borders(self, rect):
        """Paint cell borders"""

        clip_path = QPainterPath()  # Required for clipping in SVG export
        clip_path.addRect(rect)

        self.painter.save()

        self.paint_bottom_border(rect, clip_path)
        self.paint_right_border(rect, clip_path)
        self.paint_above_borders(rect, clip_path)
        self.paint_left_borders(rect, clip_path)

        self.paint_top_left_edge(rect, clip_path)
        self.paint_top_right_edge(rect, clip_path)
        self.paint_bottom_left_edge(rect, clip_path)
        self.paint_bottom_right_edge(rect, clip_path)

        self.painter.restore()

    def paint(self):
        """Paints the cell"""

        rect = QRectF(self.option.rect)

        with painter_save(self.painter):
            self.painter.setClipRect(self.option.rect)

            angle = self.cell_attributes[self.key].angle
            inner_rect = self.inner_rect(rect)

            with painter_rotate(self.painter, inner_rect, angle) as rrect:
                self.paint_content(rrect)

            self.paint_borders(rect)



fig = Figure()
ax = fig.add_axes([.2,.2, .7, .7])

N_points = 100000
n_bins = 20

# Fixing random state for reproducibility
numpy.random.seed(19680801)

# Generate a normal distribution, center at x=0 and y=5
x = numpy.random.randn(N_points)
y = .4 * x + numpy.random.randn(100000) + 5

# We can set the number of bins with the `bins` kwarg
# We can also define custom numbers of bins for each axis
ax.hist2d(x, y, bins=(50,30))

ax.set(xlabel='X', ylabel='Y')
ax.set_title('Matrix chart', pad=20)

fig



# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Copyright Seongyong Park
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

"""
parsers
======

Provides parsers to apply on expressions that are not pure Python expressions

Provides
--------

* spreadsheet_formula_to_code: Spreadsheet formula parser

"""

try:
    import pycel
    from pycel import excelformula
except ImportError:
    pycel = None


def spreadsheet_formula_to_code(formula: str) -> str:
    """ Converts spreadsheet formula into Python code via pycel """
    if pycel is None:
        return formula

    ex = excelformula.ExcelFormula(formula)
    return ex.python_code



# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

"""

High resolution svg support for qimage and :mod:`matplotlib.figure`

**Provides**

* :class:`QImageSvg`


"""

from io import StringIO

from PyQt6.QtGui import QImage, QPainter
try:
    from PyQt6.QtSvg import QSvgRenderer
except ImportError:
    QSvgRenderer = None

try:
    from matplotlib.figure import Figure
    from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
except ImportError:
    Figure = None


class QImageSvg(QImage):
    """Subclass of PyQt6.QtGui

    Adds support for SVG byte strings and matplotlib figures

    """

    def from_svg_bytes(self, svg_bytes: bytes):
        """Paints an svg from a bytes object

        :param svg_bytes: SVG file content

        """

        renderer = QSvgRenderer(svg_bytes)
        painter = QPainter(self)
        painter.eraseRect(self.rect())
        renderer.render(painter)
        painter.end()

    def _matplotlib_figure2svg_bytes(self, figure: Figure) -> bytes:
        """Converts a a matplotlib figure to an SVG bytes object

        :param figure: Matplotib figure
        :return: SVG file content

        """

        canvas = FigureCanvasQTAgg(figure)
        svg_filelike = StringIO()
        figure.savefig(svg_filelike, format="svg")
        svg_filelike.seek(0)
        svg_bytes = bytes(svg_filelike.read(), encoding='utf-8')
        svg_filelike.close()

        return svg_bytes

    def from_matplotlib(self, figure: Figure):
        """Paints an svg from a matplotlib figure

        :param figure: Matplotib figure

        """

        if not isinstance(figure, Figure):
            msg_tpl = "figure must be instance of {}."
            msg = msg_tpl.format(Figure)
            raise ValueError(msg)

        svg_bytes = self._matplotlib_figure2svg_bytes(figure)
        self.from_svg_bytes(svg_bytes)
